"""
Utility functions for the UEA multivariate time series classification
archive.
"""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import os
import pickle
from scipy.io.arff import loadarff
from sklearn.utils import Bunch
from urllib.request import urlretrieve
import zipfile


def _correct_uea_name_download(dataset):
    if dataset == 'Ering':
        return 'ERing'
    else:
        return dataset


def uea_dataset_list():
    """List of available UEA datasets.

    Returns
    -------
    datasets : list
        List of available datasets from the UEA Time Series
        Classification Archive.

    References
    ----------
    .. [1] `List of datasets on the UEA & UCR archive
           <http://www.timeseriesclassification.com/dataset.php>`_

    Examples
    --------
    >>> from pyts.datasets import uea_dataset_list
    >>> uea_dataset_list()[:3]
    ['ArticularyWordRecognition', 'AtrialFibrillation', 'BasicMotions']

    """
    module_path = os.path.dirname(__file__)
    finfo = os.path.join(module_path, 'info', 'uea.pickle')
    dictionary = pickle.load(open(finfo, 'rb'))
    datasets = sorted(dictionary.keys())
    return datasets


def uea_dataset_info(dataset=None):
    """Information about the UEA datasets.

    Parameters
    ----------
    dataset : str, list of str or None (default = None)
        The data sets for which the information will be returned.
        If None, the information for all the datasets is returned.

    Returns
    -------
    dictionary : dict
        Dictionary with the information for each dataset.

    References
    ----------
    .. [1] `List of datasets on the UEA & UCR archive
           <http://www.timeseriesclassification.com/dataset.php>`_

    Examples
    --------
    >>> from pyts.datasets import uea_dataset_info
    >>> uea_dataset_info('AtrialFibrillation')['n_timestamps']
    640

    """
    module_path = os.path.dirname(__file__)
    finfo = os.path.join(module_path, 'info', 'uea.pickle')
    dictionary = pickle.load(open(finfo, 'rb'))
    datasets = list(dictionary.keys())

    if dataset is None:
        return dictionary
    elif isinstance(dataset, str):
        if dataset not in datasets:
            raise ValueError(
                "{0} is not a valid name. The list of available names "
                "can be obtained by calling the "
                "'pyts.datasets.uea_dataset_list' function."
                .format(dataset)
            )
        else:
            return dictionary[dataset]
    elif isinstance(dataset, (list, tuple, np.ndarray)):
        dataset = np.asarray(dataset)
        invalid_datasets = np.setdiff1d(dataset, datasets)
        if invalid_datasets.size > 0:
            raise ValueError(
                "The following names are not valid: {0}. The list of "
                "available names can be obtained by calling the "
                "'pyts.datasets.uea_dataset_list' function."
                .format(invalid_datasets)
            )
        else:
            info = {}
            for data in dataset:
                info[data] = dictionary[data]
            return info


def fetch_uea_dataset(dataset, use_cache=True, data_home=None,
                      return_X_y=False):  # noqa 207
    """Fetch dataset from UEA TSC Archive by name.

    Fetched data sets are saved by default in the
    ``pyts/datasets/cached_datasets/UEA/`` folder. To avoid
    downloading the same data set several times, it is
    highly recommended not to change the default values
    of ``use_cache`` and ``path``.

    Parameters
    ----------
    dataset : str
        Name of the dataset.

    use_cache : bool (default = True)
        If True, look if the data set has already been fetched
        and load the fetched version if it is the case. If False,
        download the data set from the UCR Time Series Classification
        Archive.

    data_home : None or str (default = None)
        The path of the folder containing the cached data set.
        If None, the ``pyts.datasets.cached_datasets/UEA/`` folder is
        used. If the data set is not found, it is downloaded and cached
        in this path.

    return_X_y : bool (default = False)
        If True, returns ``(data_train, data_test, target_train, target_test)``
        instead of a Bunch object. See below for more information about the
        `data` and `target` object.

    Returns
    -------
    data : Bunch
        Dictionary-like object, with attributes:

        data_train : array of floats
            The time series in the training set.
        data_test : array of floats
            The time series in the test set.
        target_train : array of integers
            The classification labels in the training set.
        target_test : array of integers
            The classification labels in the test set.
        DESCR : str
            The full description of the dataset.
        url : str
            The url of the dataset.

    (data_train, data_test, target_train, target_test) : tuple if \
``return_X_y`` is True

    Notes
    -----
    Missing values are represented as NaN's.

    References
    ----------
    .. [1] A. Bagnall et al, "The UEA multivariate time series
           classification archive, 2018". arXiv:1811.00075 [cs, stat],
           2018.

    .. [2] A. Bagnall et al, "The UEA & UCR Time Series Classification
           Repository", www.timeseriesclassification.com.

    """
    if dataset not in uea_dataset_list():
        raise ValueError(
            "{0} is not a valid name. The list of available names "
            "can be obtained with ``pyts.datasets.uea_dataset_list()``"
            .format(dataset)
        )
    if data_home is None:
        import pyts
        home = '/'.join(pyts.__file__.split('/')[:-2]) + '/'
        relative_path = 'pyts/datasets/cached_datasets/UEA/'
        path = home + relative_path
    else:
        path = data_home
    if not os.path.exists(path):
        os.makedirs(path)

    correct_dataset = _correct_uea_name_download(dataset)
    if use_cache and os.path.exists(path + correct_dataset):
        bunch = _load_uea_dataset(correct_dataset, path)
    else:
        url = ("http://www.timeseriesclassification.com/Downloads/{0}.zip"
               .format(correct_dataset))
        filename = 'temp_{}'.format(correct_dataset)
        _ = urlretrieve(url, path + filename)
        zipfile.ZipFile(path + filename).extractall(path + correct_dataset)
        os.remove(path + filename)
        bunch = _load_uea_dataset(correct_dataset, path)

    if return_X_y:
        return (bunch.data_train, bunch.data_test,
                bunch.target_train, bunch.target_test)
    return bunch


def _load_uea_dataset(dataset, path):
    """Load a UEA data set from a local folder.

    Parameters
    ----------
    dataset : str
        Name of the dataset.

    path : str
        The path of the folder containing the cached data set.

    Returns
    -------
    data : Bunch
        Dictionary-like object, with attributes:

        data_train : array of floats
            The time series in the training set.
        data_test : array of floats
            The time series in the test set.
        target_train : array
            The classification labels in the training set.
        target_test : array
            The classification labels in the test set.
        DESCR : str
            The full description of the dataset.
        url : str
            The url of the dataset.

    Notes
    -----
    Missing values are represented as NaN's.

    """
    new_path = path + dataset + '/'
    try:
        description_file = [
            file for file in os.listdir(new_path)
            if ('Description.txt' in file
                or dataset + '.txt' in file)
        ][0]
    except IndexError:
        description_file = None

    if description_file is not None:
        try:
            with(open(new_path + description_file, encoding='utf-8')) as f:
                description = f.read()
        except UnicodeDecodeError:
            with(open(new_path + description_file,
                      encoding='ISO-8859-1')) as f:
                description = f.read()
    else:
        description = None

    data_train = loadarff(new_path + dataset + '_TRAIN.arff')
    X_train, y_train = _parse_relational_arff(data_train)

    data_test = loadarff(new_path + dataset + '_TEST.arff')
    X_test, y_test = _parse_relational_arff(data_test)

    bunch = Bunch(
        data_train=X_train, target_train=y_train,
        data_test=X_test, target_test=y_test,
        DESCR=description,
        url=("http://www.timeseriesclassification.com/"
             "description.php?Dataset={}".format(dataset))
    )

    return bunch


def _parse_relational_arff(data):
    X_data = np.asarray(data[0])
    n_samples = len(X_data)
    X, y = [], []

    if X_data[0][0].dtype.names is None:
        for i in range(n_samples):
            X_sample = np.asarray(
                [X_data[i][name] for name in X_data[i].dtype.names]
            )
            X.append(X_sample.T)
            y.append(X_data[i][1])
    else:
        for i in range(n_samples):
            X_sample = np.asarray(
                [X_data[i][0][name] for name in X_data[i][0].dtype.names]
            )
            X.append(X_sample.T)
            y.append(X_data[i][1])

    X = np.asarray(X).astype('float64')
    y = np.asarray(y)

    try:
        y = y.astype('float64').astype('int64')
    except ValueError:
        y = y.astype(str)

    return X, y



"""Testing for MultivariateTransformer."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
import re
from scipy.sparse import csr_matrix
from pyts.classification import SAXVSM
from pyts.image import RecurrencePlot
from pyts.multivariate.transformation import MultivariateTransformer
from pyts.transformation import BOSS


n_samples, n_features, n_timestamps = 40, 3, 30
rng = np.random.RandomState(42)
X = rng.randn(n_samples, n_features, n_timestamps)


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'estimator': [BOSS(), RecurrencePlot(), SAXVSM()]},
      ValueError, "Estimator 2 must be a transformer."),

     ({'estimator': [BOSS()]}, ValueError,
      "If 'estimator' is a list, its length must be equal to "
      "the number of features (1 != 3)"),

     ({'estimator': None}, TypeError,
      "'estimator' must be a transformer that inherits from "
      "sklearn.base.BaseEstimator or a list thereof.")]
)
def test_parameter_check(params, error, err_msg):
    """Test parameter validation."""
    transformer = MultivariateTransformer(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        transformer.fit_transform(X)


@pytest.mark.parametrize(
    'X, arr_desired',
    [(csr_matrix(np.ones((5, 5))), np.ones((5, 5))),
     (np.ones((5, 5)), np.ones((5, 5)))]
)
def test_array_conversion(X, arr_desired):
    """Test the array conversion static method."""
    arr_actual = MultivariateTransformer._convert_to_array(X)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'X, err_msg',
    [({}, "Unexpected type for X: dict."),
     ({0}, "Unexpected type for X: set."),
     ([], "Unexpected type for X: list.")]
)
def test_array_conversion_error(X, err_msg):
    """Test the array conversion static method."""
    with pytest.raises(ValueError, match=re.escape(err_msg)):
        MultivariateTransformer._convert_to_array(X)


@pytest.mark.parametrize(
    'params, shape_desired',
    [({'estimator': RecurrencePlot(dimension=1), 'flatten': False},
      (40, 3, 30, 30)),

     ({'estimator': RecurrencePlot(dimension=6), 'flatten': False},
      (40, 3, 25, 25)),

     ({'estimator': RecurrencePlot(dimension=6), 'flatten': True},
      (40, 3 * 25 * 25)),

     ({'estimator': [RecurrencePlot(dimension=6),
                     RecurrencePlot(dimension=4),
                     RecurrencePlot(dimension=2)], 'flatten': True},
      (40, (25 * 25) + (27 * 27) + (29 * 29)))]
)
def test_shapes(params, shape_desired):
    """Test that the shape of the output is the expected one."""
    transformer = MultivariateTransformer(**params)
    assert transformer.fit(X).transform(X).shape == shape_desired
    assert transformer.fit_transform(X).shape == shape_desired


@pytest.mark.parametrize(
    'params, ndim_desired',
    [({'estimator': RecurrencePlot(), 'flatten': False}, 4),
     ({'estimator': RecurrencePlot(dimension=6), 'flatten': True}, 2),
     ({'estimator': BOSS(), 'flatten': False}, 2),
     ({'estimator': BOSS(), 'flatten': True}, 2),
     ({'estimator': [RecurrencePlot(dimension=6),
                     RecurrencePlot(dimension=4),
                     RecurrencePlot(dimension=2)], 'flatten': True}, 2),
     ({'estimator': [RecurrencePlot(dimension=6),
                     RecurrencePlot(dimension=4),
                     RecurrencePlot(dimension=2)], 'flatten': False}, 2),
     ({'estimator': [RecurrencePlot(dimension=6), BOSS(),
                     RecurrencePlot(dimension=2)], 'flatten': True}, 2)]
)
def test_ndim(params, ndim_desired):
    """Test that the ndim of the output is the expected one."""
    transformer = MultivariateTransformer(**params)
    assert transformer.fit(X).transform(X).ndim == ndim_desired
    assert transformer.fit_transform(X).ndim == ndim_desired


def test_actual_results_without_flatten():
    """Test that the actual results are the expected ones."""
    params = {'estimator': RecurrencePlot(dimension=6), 'flatten': False}
    arr_actual = MultivariateTransformer(**params).fit_transform(X)
    arr_desired = []
    for i in range(n_features):
        arr_desired.append(params['estimator'].transform(X[:, i]))
    arr_desired = np.transpose(arr_desired, axes=(1, 0, 2, 3))
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


def test_actual_results_with_flatten():
    """Test that the actual results are the expected ones."""
    params = {'estimator': RecurrencePlot(dimension=6), 'flatten': True}
    arr_actual = MultivariateTransformer(**params).fit_transform(X)
    arr_desired = []
    for i in range(n_features):
        arr_desired.append(params['estimator'].transform(X[:, i]).reshape(
            (n_samples, -1)))
    arr_desired = np.concatenate(arr_desired, axis=1)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)



"""The :mod:`pyts.approximation` module includes approximation algorithms."""

from .paa import PiecewiseAggregateApproximation
from .sax import SymbolicAggregateApproximation
from .dft import DiscreteFourierTransform
from .mcb import MultipleCoefficientBinning
from .sfa import SymbolicFourierApproximation


__all__ = ['PiecewiseAggregateApproximation',
           'SymbolicAggregateApproximation',
           'DiscreteFourierTransform',
           'MultipleCoefficientBinning',
           'SymbolicFourierApproximation']



"""Code for Markov Transition Field."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
from math import ceil
from numba import njit, prange
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_array
from ..base import UnivariateTransformerMixin
from ..preprocessing import KBinsDiscretizer
from ..utils import segmentation


@njit()
def _markov_transition_matrix(X_binned, n_samples, n_timestamps, n_bins):
    X_mtm = np.zeros((n_samples, n_bins, n_bins))
    for i in prange(n_samples):
        for j in prange(n_timestamps - 1):
            X_mtm[i, X_binned[i, j], X_binned[i, j + 1]] += 1
    return X_mtm


@njit()
def _markov_transition_field(
    X_binned, X_mtm, n_samples, n_timestamps, n_bins
):
    X_mtf = np.zeros((n_samples, n_timestamps, n_timestamps))
    for i in prange(n_samples):
        for j in prange(n_timestamps):
            for k in prange(n_timestamps):
                X_mtf[i, j, k] = X_mtm[i, X_binned[i, j], X_binned[i, k]]
    return X_mtf


@njit()
def _aggregated_markov_transition_field(X_mtf, n_samples, image_size,
                                        start, end):
    X_amtf = np.empty((n_samples, image_size, image_size))
    for i in prange(n_samples):
        for j in prange(image_size):
            for k in prange(image_size):
                X_amtf[i, j, k] = np.mean(
                    X_mtf[i, start[j]:end[j], start[k]:end[k]]
                )
    return X_amtf


class MarkovTransitionField(BaseEstimator, UnivariateTransformerMixin):
    """Markov Transition Field.

    Parameters
    ----------
    image_size : int or float (default = 1.)
        Shape of the output images. If float, it represents a percentage
        of the size of each time series and must be between 0 and 1. Output
        images are square, thus providing the size of one dimension is enough.

    n_bins : int (default = 5)
        Number of bins (also known as the size of the alphabet)

    strategy : 'uniform', 'quantile' or 'normal' (default = 'quantile')
        Strategy used to define the widths of the bins:

        - 'uniform': All bins in each sample have identical widths
        - 'quantile': All bins in each sample have the same number of points
        - 'normal': Bin edges are quantiles from a standard normal distribution


    overlapping : bool (default = False)
        If False, reducing the image with the blurring kernel
        will be applied on non-overlapping rectangles. If True,
        it will be applied on possibly overlapping squares.

    flatten : bool (default = False)
        If True, images are flattened to be one-dimensional.

    References
    ----------
    .. [1] Z. Wang and T. Oates, "Encoding Time Series as Images for Visual
           Inspection and Classification Using Tiled Convolutional Neural
           Networks." AAAI Workshop (2015).

    Examples
    --------
    >>> from pyts.datasets import load_gunpoint
    >>> from pyts.image import MarkovTransitionField
    >>> X, _, _, _ = load_gunpoint(return_X_y=True)
    >>> transformer = MarkovTransitionField()
    >>> X_new = transformer.transform(X)
    >>> X_new.shape
    (50, 150, 150)

    """

    def __init__(self, image_size=1., n_bins=8,
                 strategy='quantile', overlapping=False, flatten=False):
        self.image_size = image_size
        self.n_bins = n_bins
        self.strategy = strategy
        self.overlapping = overlapping
        self.flatten = flatten

    def fit(self, X=None, y=None):
        """Pass.

        Parameters
        ----------
        X
            Ignored
        y
            Ignored

        Returns
        -------
        self : object

        """
        return self

    def transform(self, X):
        """Transform each time series into a MTF image.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Input data

        Returns
        -------
        X_new : array-like, shape = (n_samples, image_size, image_size)
            Transformed data. If ``flatten=True``, the shape is
            `(n_samples, image_size * image_size)`.

        """
        X = check_array(X)
        n_samples, n_timestamps = X.shape
        image_size = self._check_params(n_timestamps)

        discretizer = KBinsDiscretizer(n_bins=self.n_bins,
                                       strategy=self.strategy)
        X_binned = discretizer.fit_transform(X)

        X_mtm = _markov_transition_matrix(X_binned, n_samples,
                                          n_timestamps, self.n_bins)
        sum_mtm = X_mtm.sum(axis=2)
        np.place(sum_mtm, sum_mtm == 0, 1)
        X_mtm /= sum_mtm[:, :, None]

        X_mtf = _markov_transition_field(
            X_binned, X_mtm, n_samples, n_timestamps, self.n_bins
        )

        window_size, remainder = divmod(n_timestamps, image_size)
        if remainder == 0:
            X_amtf = np.reshape(
                X_mtf, (n_samples, image_size, window_size,
                        image_size, window_size)
            ).mean(axis=(2, 4))
        else:
            window_size += 1
            start, end, _ = segmentation(
                n_timestamps, window_size, self.overlapping, image_size
            )
            X_amtf = _aggregated_markov_transition_field(
                X_mtf, n_samples, image_size, start, end
            )
        if self.flatten:
            return X_amtf.reshape(n_samples, -1)
        return X_amtf

    def _check_params(self, n_timestamps):
        if not isinstance(self.image_size,
                          (int, np.integer, float, np.floating)):
            raise TypeError("'image_size' must be an integer or a float.")
        if isinstance(self.image_size, (int, np.integer)):
            if self.image_size < 1 or self.image_size > n_timestamps:
                raise ValueError(
                    "If 'image_size' is an integer, it must be greater "
                    "than or equal to 1 and lower than or equal to "
                    "n_timestamps (got {0}).".format(self.image_size)
                )
            image_size = self.image_size
        else:
            if self.image_size < 0. or self.image_size > 1.:
                raise ValueError(
                    "If 'image_size' is a float, it must be greater "
                    "than 0 and lower than or equal to 1 (got {0})."
                    .format(self.image_size)
                )
            image_size = ceil(self.image_size * n_timestamps)
        if not isinstance(self.n_bins, (int, np.integer)):
            raise TypeError("'n_bins' must be an integer.")
        if not self.n_bins >= 2:
            raise ValueError("'n_bins' must be greater than or equal to 2.")
        if self.strategy not in ['uniform', 'quantile', 'normal']:
            raise ValueError("'strategy' must be 'uniform', 'quantile' or "
                             "'normal'.")
        return image_size



"""Testing for base classes."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
from sklearn.base import clone

from pyts.classification import SAXVSM
from pyts.datasets import load_gunpoint, load_basic_motions
from pyts.multivariate.image import JointRecurrencePlot
from pyts.multivariate.classification import MultivariateClassifier
from pyts.approximation import SymbolicFourierApproximation


X_uni, _, y_uni, _ = load_gunpoint(return_X_y=True)
X_multi, _, y_multi, _ = load_basic_motions(return_X_y=True)


@pytest.mark.parametrize(
    'estimator, X, y',
    [(SymbolicFourierApproximation(n_bins=2), X_uni, None),
     (SymbolicFourierApproximation(n_bins=2, strategy='entropy'),
      X_uni, y_uni)]
)
def test_univariate_transformer_mixin(estimator, X, y):
    sfa_1 = clone(estimator)
    sfa_2 = clone(estimator)
    np.testing.assert_array_equal(
        sfa_1.fit_transform(X, y), sfa_2.fit(X, y).transform(X)
    )


@pytest.mark.parametrize(
    'estimator, X, y',
    [(JointRecurrencePlot(), X_multi, None),
     (JointRecurrencePlot(), X_multi, y_multi)]
)
def test_multivariate_transformer_mixin(estimator, X, y):
    jrp_1 = clone(estimator)
    jrp_2 = clone(estimator)
    np.testing.assert_allclose(
        jrp_1.fit_transform(X, y), jrp_2.fit(X, y).transform(X)
    )


@pytest.mark.parametrize(
    'sample_weight',
    [None, np.ones_like(y_uni), np.random.uniform(size=y_uni.size)]
)
def test_univariate_classifier_mixin(sample_weight):
    clf = SAXVSM().fit(X_uni, y_uni)
    assert isinstance(clf.score(X_uni, y_uni, sample_weight),
                      (float, np.floating))


@pytest.mark.parametrize(
    'sample_weight',
    [None, np.ones(y_multi.size), np.random.uniform(size=y_multi.size)]
)
def test_multivariate_classifier_mixin(sample_weight):
    clf = MultivariateClassifier(SAXVSM()).fit(X_multi, y_multi)
    assert isinstance(clf.score(X_multi, y_multi, sample_weight),
                      (float, np.floating))



"""Code for Bag-of-SFA Symbols."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
from math import ceil
from scipy.sparse import csr_matrix
from sklearn.base import BaseEstimator
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.utils.validation import check_array, check_is_fitted
from sklearn.utils.multiclass import check_classification_targets
from ..approximation import SymbolicFourierApproximation
from ..base import UnivariateTransformerMixin
from ..utils.utils import _windowed_view


class BOSS(BaseEstimator, UnivariateTransformerMixin):
    """Bag of Symbolic Fourier Approximation Symbols.

    For each time series, subseries are extracted using a slidind window.
    Then the subseries are transformed into a word using the Symbolic
    Fourier Approximation (SFA) algorithm. For each time series, the words
    are grouped together and a histogram counting the occurences of each
    word is created.

    Parameters
    ----------
    word_size : int (default = 4)
        Size of each word.

    n_bins : int (default = 4)
        The number of bins to produce. It must be between 2 and 26.

    strategy : str (default = 'quantile')
        Strategy used to define the widths of the bins:

        - 'uniform': All bins in each sample have identical widths
        - 'quantile': All bins in each sample have the same number of points
        - 'normal': Bin edges are quantiles from a standard normal distribution
        - 'entropy': Bin edges are computed using information gain

    window_size : int or float (default = 10)
        Size of the sliding window. If float, it represents the percentage of
        the size of each time series and must be between 0 and 1. The window
        size will be computed as ``ceil(window_size * n_timestamps)``.

    window_step : int or float (default = 1)
        Step of the sliding window. If float, it represents the percentage of
        the size of each time series and must be between 0 and 1. The window
        size will be computed as ``ceil(window_step * n_timestamps)``.

    anova : bool (default = False)
        If True, the Fourier coefficient selection is done via a one-way
        ANOVA test. If False, the first Fourier coefficients are selected.

    drop_sum : bool (default = False)
        If True, the first Fourier coefficient (i.e. the sum of the subseries)
        is dropped. Otherwise, it is kept.

    norm_mean : bool (default = False)
        If True, center each subseries before scaling.

    norm_std : bool (default = False)
        If True, scale each subseries to unit variance.

    numerosity_reduction : bool (default = True)
        If True, delete sample-wise all but one occurence of back to back
        identical occurences of the same words.

    sparse : bool (default = True)
        Return a sparse matrix if True, else return an array.

    alphabet : None, 'ordinal' or array-like, shape = (n_bins,)
        Alphabet to use. If None, the first `n_bins` letters of the Latin
        alphabet are used.

    Attributes
    ----------
    vocabulary_ : dict
        A mapping of feature indices to terms.

    References
    ----------
    .. [1] P. Schäfer, "The BOSS is concerned with time series classification
           in the presence of noise". Data Mining and Knowledge Discovery,
           29(6), 1505-1530 (2015).

    Examples
    --------
    >>> from pyts.datasets import load_gunpoint
    >>> from pyts.transformation import BOSS
    >>> X_train, X_test, _, _ = load_gunpoint(return_X_y=True)
    >>> boss = BOSS(word_size=2, n_bins=2, sparse=False)
    >>> boss.fit(X_train)
    BOSS(...)
    >>> sorted(boss.vocabulary_.values())
    ['aa', 'ab', 'ba', 'bb']
    >>> boss.transform(X_test)
    array(...)

    """

    def __init__(self, word_size=4, n_bins=4, strategy='quantile',
                 window_size=10, window_step=1, anova=False, drop_sum=False,
                 norm_mean=False, norm_std=False, numerosity_reduction=True,
                 sparse=True, alphabet=None):
        self.word_size = word_size
        self.n_bins = n_bins
        self.strategy = strategy
        self.window_size = window_size
        self.window_step = window_step
        self.anova = anova
        self.drop_sum = drop_sum
        self.norm_mean = norm_mean
        self.norm_std = norm_std
        self.numerosity_reduction = numerosity_reduction
        self.sparse = sparse
        self.alphabet = alphabet

    def fit(self, X, y=None):
        """Fit the model according to the given training data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Training vector.

        y : None or array-like, shape = (n_samples,)
            Class labels for each data sample.

        Returns
        -------
        self : object

        """
        X = check_array(X)
        n_samples, n_timestamps = X.shape
        if y is not None:
            check_classification_targets(y)

        window_size, window_step = self._check_params(n_timestamps)
        n_windows = (n_timestamps - window_size + window_step) // window_step

        X_windowed = _windowed_view(
            X, n_samples, n_timestamps, window_size, window_step
        )
        X_windowed = X_windowed.reshape(n_samples * n_windows, window_size)

        sfa = SymbolicFourierApproximation(
            n_coefs=self.word_size, drop_sum=self.drop_sum, anova=self.anova,
            norm_mean=self.norm_mean, norm_std=self.norm_std,
            n_bins=self.n_bins, strategy=self.strategy, alphabet=self.alphabet
        )
        if y is None:
            y_repeated = None
        else:
            y_repeated = np.repeat(y, n_windows)
        X_sfa = sfa.fit_transform(X_windowed, y_repeated)

        X_word = np.asarray([''.join(X_sfa[i])
                             for i in range(n_samples * n_windows)])
        X_word = X_word.reshape(n_samples, n_windows)

        if self.numerosity_reduction:
            not_equal = np.c_[X_word[:, 1:] != X_word[:, :-1],
                              np.full(n_samples, True)]
            X_bow = np.asarray([' '.join(X_word[i, not_equal[i]])
                                for i in range(n_samples)])
        else:
            X_bow = np.asarray([' '.join(X_word[i]) for i in range(n_samples)])

        vectorizer = CountVectorizer()
        vectorizer.fit(X_bow)
        self.vocabulary_ = {value: key for key, value in
                            vectorizer.vocabulary_.items()}
        self._window_size = window_size
        self._window_step = window_step
        self._n_windows = n_windows
        self._sfa = sfa
        self._vectorizer = vectorizer
        return self

    def transform(self, X):
        """Transform the provided data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Test samples.

        Returns
        -------
        X_new : sparse matrix, shape = (n_samples, n_words)
            Document-term matrix.

        """
        check_is_fitted(self, ['_sfa', '_vectorizer', 'vocabulary_'])
        X = check_array(X)
        n_samples, n_timestamps = X.shape

        X_windowed = _windowed_view(
            X, n_samples, n_timestamps, self._window_size, self._window_step
        )
        X_windowed = X_windowed.reshape(-1, self._window_size)

        X_sfa = self._sfa.transform(X_windowed)
        X_word = np.asarray([''.join(X_sfa[i]) for i in range(X_sfa.shape[0])])
        X_word = X_word.reshape(n_samples, self._n_windows)

        if self.numerosity_reduction:
            not_equal = np.c_[X_word[:, 1:] != X_word[:, :-1],
                              np.full(n_samples, True)]
            X_bow = np.asarray([' '.join(X_word[i, not_equal[i]])
                                for i in range(n_samples)])
        else:
            X_bow = np.asarray([' '.join(X_word[i]) for i in range(n_samples)])

        X_boss = self._vectorizer.transform(X_bow)
        if not self.sparse:
            return X_boss.A
        return csr_matrix(X_boss)

    def fit_transform(self, X, y=None):
        """Fit the data then transform it.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Training vector.

        y : None or array-like, shape = (n_samples,)
            Class labels for each data sample.

        Returns
        -------
        X_new : sparse matrix, shape = (n_samples, n_words)
            Document-term matrix.

        """
        X = check_array(X)
        n_samples, n_timestamps = X.shape
        if y is not None:
            check_classification_targets(y)

        window_size, window_step = self._check_params(n_timestamps)
        n_windows = (n_timestamps - window_size + window_step) // window_step

        X_windowed = _windowed_view(
            X, n_samples, n_timestamps, window_size, window_step
        )
        X_windowed = X_windowed.reshape(n_samples * n_windows, window_size)

        sfa = SymbolicFourierApproximation(
            n_coefs=self.word_size, drop_sum=self.drop_sum, anova=self.anova,
            norm_mean=self.norm_mean, norm_std=self.norm_std,
            n_bins=self.n_bins, strategy=self.strategy, alphabet=self.alphabet
        )
        if y is None:
            y_repeated = None
        else:
            y_repeated = np.repeat(y, n_windows)
        X_sfa = sfa.fit_transform(X_windowed, y_repeated)

        X_word = np.asarray([''.join(X_sfa[i])
                             for i in range(n_samples * n_windows)])
        X_word = X_word.reshape(n_samples, n_windows)

        if self.numerosity_reduction:
            not_equal = np.c_[X_word[:, 1:] != X_word[:, :-1],
                              np.full(n_samples, True)]
            X_bow = np.asarray([' '.join(X_word[i, not_equal[i]])
                                for i in range(n_samples)])
        else:
            X_bow = np.asarray([' '.join(X_word[i]) for i in range(n_samples)])

        vectorizer = CountVectorizer()
        X_boss = vectorizer.fit_transform(X_bow)
        self.vocabulary_ = {value: key for key, value in
                            vectorizer.vocabulary_.items()}
        self._window_size = window_size
        self._window_step = window_step
        self._n_windows = n_windows
        self._sfa = sfa
        self._vectorizer = vectorizer
        if not self.sparse:
            return X_boss.A
        return csr_matrix(X_boss)

    def _check_params(self, n_timestamps):
        if not isinstance(self.word_size, (int, np.integer)):
            raise TypeError("'word_size' must be an integer.")
        if not self.word_size >= 1:
            raise ValueError("'word_size' must be a positive integer.")

        if not isinstance(self.window_size,
                          (int, np.integer, float, np.floating)):
            raise TypeError("'window_size' must be an integer or a float.")
        if isinstance(self.window_size, (int, np.integer)):
            if self.drop_sum:
                if not 1 <= self.window_size <= (n_timestamps - 1):
                    raise ValueError(
                        "If 'window_size' is an integer, it must be greater "
                        "than or equal to 1 and lower than or equal to "
                        "(n_timestamps - 1) if 'drop_sum=True'."
                    )
            else:
                if not 1 <= self.window_size <= n_timestamps:
                    raise ValueError(
                        "If 'window_size' is an integer, it must be greater "
                        "than or equal to 1 and lower than or equal to "
                        "n_timestamps if 'drop_sum=False'."
                    )
            window_size = self.window_size
        else:
            if not 0 < self.window_size <= 1:
                raise ValueError(
                    "If 'window_size' is a float, it must be greater "
                    "than 0 and lower than or equal to 1."
                )
            window_size = ceil(self.window_size * n_timestamps)

        if not isinstance(self.window_step,
                          (int, np.integer, float, np.floating)):
            raise TypeError("'window_step' must be an integer or a float.")
        if isinstance(self.window_step, (int, np.integer)):
            if not 1 <= self.window_step <= n_timestamps:
                raise ValueError(
                    "If 'window_step' is an integer, it must be greater "
                    "than or equal to 1 and lower than or equal to "
                    "n_timestamps."
                )
            window_step = self.window_step
        else:
            if not 0 < self.window_step <= 1:
                raise ValueError(
                    "If 'window_step' is a float, it must be greater "
                    "than 0 and lower than or equal to 1."
                )
            window_step = ceil(self.window_step * n_timestamps)
        if self.drop_sum:
            if not self.word_size <= (window_size - 1):
                raise ValueError(
                    "'word_size' must be lower than or equal to "
                    "(window_size - 1) if 'drop_sum=True'."
                )
        else:
            if not self.word_size <= window_size:
                raise ValueError(
                    "'word_size' must be lower than or equal to "
                    "window_size if 'drop_sum=False'."
                )
        return window_size, window_step



"""Testing for Bag-of-Words."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
import re
from pyts.bag_of_words import BagOfWords, WordExtractor

# ######################### Tests for WordExtractor #########################

X_word = [['a', 'a', 'a', 'b', 'a'],
          ['a', 'a', 'b', 'b', 'a'],
          ['b', 'b', 'b', 'b', 'a']]


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'window_size': '4'}, TypeError,
      "'window_size' must be an integer or a float."),
     ({'window_step': [0, 1]}, TypeError,
      "'window_step' must be an integer or a float."),
     ({'window_size': 0}, ValueError,
      "If 'window_size' is an integer, it must be greater than or equal to 1 "
      "and lower than or equal to n_timestamps (got 0)."),
     ({'window_size': 2.}, ValueError,
      "If 'window_size' is a float, it must be greater than 0 and lower "
      "than or equal to 1 (got {0}).".format(2.)),
     ({'window_step': 0}, ValueError,
      "If 'window_step' is an integer, it must be greater than or equal to 1 "
      "and lower than or equal to n_timestamps (got 0)."),
     ({'window_step': 2.}, ValueError,
      "If 'window_step' is a float, it must be greater than 0 and lower "
      "than or equal to 1 (got {0}).".format(2.))]
)
def test_parameter_check_word_extractor(params, error, err_msg):
    """Test parameter validation."""
    bow = WordExtractor(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        bow.transform(X_word)


@pytest.mark.parametrize(
    'params, arr_desired',
    [({}, ['a b a', 'a b a', 'b a']),
     ({'numerosity_reduction': False},
      ['a a a b a', 'a a b b a', 'b b b b a']),
     ({'window_size': 1, 'numerosity_reduction': False},
      ['a a a b a', 'a a b b a', 'b b b b a']),
     ({'window_size': 2, 'numerosity_reduction': False},
      ['aa aa ab ba', 'aa ab bb ba', 'bb bb bb ba']),
     ({'window_size': 0.4, 'window_step': 0.2, 'numerosity_reduction': False},
      ['aa aa ab ba', 'aa ab bb ba', 'bb bb bb ba']),
     ({'window_size': 2, 'window_step': 1, 'numerosity_reduction': True},
      ['aa ab ba', 'aa ab bb ba', 'bb ba']),
     ({'window_size': 3, 'window_step': 2, 'numerosity_reduction': False},
      ['aaa aba', 'aab bba', 'bbb bba']),
     ({'window_size': 3, 'window_step': 2, 'numerosity_reduction': True},
      ['aaa aba', 'aab bba', 'bbb bba'])]
)
def test_actual_results_word_extractor(params, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = WordExtractor(**params).fit_transform(X_word)
    np.testing.assert_array_equal(arr_actual, arr_desired)


# ######################### Tests for BagOfWords #########################

X_bow = np.arange(30).reshape(3, 10)
X_bow[2, :8] = 20


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'window_size': '4'}, TypeError,
      "'window_size' must be an integer or a float."),

     ({'window_size': 0}, ValueError,
      "If 'window_size' is an integer, it must be greater than or equal to 1 "
      "and lower than or equal to n_timestamps (got 0)."),

     ({'window_size': 2.}, ValueError,
      "If 'window_size' is a float, it must be greater than 0 and lower "
      "than or equal to 1 (got {0}).".format(2.)),

     ({'word_size': []}, TypeError,
      "'word_size' must be an integer or a float."),

     ({'word_size': 0}, ValueError,
      "If 'word_size' is an integer, it must be greater than or equal to 1 "
      "and lower than or equal to window_size (got 0)."),

     ({'window_size': 4, 'word_size': 5}, ValueError,
      "If 'word_size' is an integer, it must be greater than or equal to 1 "
      "and lower than or equal to window_size (got 5)."),

     ({'word_size': 2.}, ValueError,
      "If 'word_size' is a float, it must be greater than 0 and lower "
      "than or equal to 1 (got {0}).".format(2.)),

     ({'n_bins': '3'}, TypeError, "'n_bins' must be an integer."),

     ({'n_bins': 1}, ValueError,
      "'n_bins' must be greater than or equal to 2 and lower than "
      "or equal to 26 (got 1)."),

     ({'n_bins': 27}, ValueError,
      "'n_bins' must be greater than or equal to 2 and lower than "
      "or equal to 26 (got 27)."),

     ({'window_size': 6, 'word_size': 4, 'n_bins': 2, 'strategy': 'whoops'},
      ValueError,
      "'strategy' must be either 'uniform', 'quantile' or 'normal' "
      "(got {0})".format('whoops')),

     ({'window_size': 6, 'word_size': 4, 'n_bins': 2, 'window_step': [0, 1]},
      TypeError,
      "'window_step' must be an integer or a float."),

     ({'window_size': 6, 'word_size': 4, 'n_bins': 2, 'window_step': 0},
      ValueError,
      "If 'window_step' is an integer, it must be greater than or equal to 1 "
      "and lower than or equal to n_timestamps (got 0)."),

     ({'window_size': 6, 'word_size': 4, 'n_bins': 2, 'window_step': 2.},
      ValueError,
      "If 'window_step' is a float, it must be greater than 0 and lower "
      "than or equal to 1 (got {0}).".format(2.)),

     ({'window_size': 6, 'word_size': 4, 'n_bins': 2, 'threshold_std': '[-1]'},
      TypeError,
      "'threshold_std' must be an integer, a float or None."),

     ({'window_size': 6, 'word_size': 4, 'n_bins': 2, 'threshold_std': -1.},
      ValueError,
      "If 'threshold_std' is an integer or a float, it must be non-negative "
      "(got -1.0)."),

     ({'window_size': 6, 'word_size': 4, 'n_bins': 2, 'alphabet': 'whoops'},
      TypeError,
      "'alphabet' must be None or array-like with shape (n_bins,) "
      "(got {0}).".format('whoops')),

     ({'window_size': 6, 'word_size': 4, 'n_bins': 2,
       'alphabet': ['a', 'b', 'c']}, ValueError,
      "If 'alphabet' is array-like, its shape must be equal to (n_bins,).")]
)
def test_parameter_check_bag_of_words(params, error, err_msg):
    """Test parameter validation."""
    bow = BagOfWords(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        bow.transform(X_bow)


@pytest.mark.parametrize(
    'params, X, arr_desired',
    [({'window_size': 8, 'word_size': 4}, X_bow,
      ['abcd', 'abcd', 'bbbb bbbd']),

     ({'window_size': 8, 'word_size': 4, 'numerosity_reduction': False},
      X_bow, ['abcd abcd abcd', 'abcd abcd abcd', 'bbbb bbbd bbbd']),

     ({'window_size': 8, 'word_size': 4, 'strategy': 'uniform'}, X_bow,
      ['abcd', 'abcd', 'aaaa aaad']),

     ({'window_size': 8, 'word_size': 4, 'strategy': 'quantile'}, X_bow,
      ['abcd', 'abcd', 'aaaa aaac']),

     ({'window_size': 8, 'word_size': 4, 'strategy': 'uniform',
       'threshold_std': np.inf}, X_bow,
      ['abbc abcd bbcd', 'abbc abcd bbcd', 'aaaa aaab aaad']),

     ({'window_size': 8, 'word_size': 4, 'strategy': 'quantile',
       'threshold_std': np.inf}, X_bow,
      ['abbc abcd bbcd', 'abbc abcd bbcd', 'aaaa aaab']),

     ({'window_size': 8, 'word_size': 4, 'strategy': 'uniform',
       'threshold_std': None}, X_bow,
      ['abcd', 'abcd', 'aaaa aaad']),

     ({'window_size': 8, 'word_size': 4, 'strategy': 'quantile',
       'threshold_std': None}, X_bow,
      ['abcd', 'abcd', 'aaaa aaac']),

     ({'window_size': 8, 'word_size': 4, 'strategy': 'uniform',
       'threshold_std': 2.4}, X_bow,
      ['abbc abcd bbcd', 'abbc abcd bbcd', 'aaaa aaad']),

     ({'window_size': 8, 'word_size': 4, 'strategy': 'quantile',
       'threshold_std': 2.4}, X_bow,
      ['abbc abcd bbcd', 'abbc abcd bbcd', 'aaaa aaac']),

     ({'window_size': 8, 'word_size': 4, 'alphabet': ['y', 'o', 'l', 'o']},
      X_bow, ['yolo', 'yolo', 'oooo']),

     ({'window_size': 0.5, 'word_size': 4}, X_bow[:2], ['abcd', 'abcd']),

     ({'window_size': 4, 'word_size': 1., 'numerosity_reduction': False},
      X_bow[:2], ['abcd abcd abcd abcd abcd abcd abcd',
                  'abcd abcd abcd abcd abcd abcd abcd']),

     ({'window_size': 4, 'word_size': 4, 'window_step': 2,
       'numerosity_reduction': False},
      X_bow[:2], ['abcd abcd abcd abcd', 'abcd abcd abcd abcd']),

     ({'window_size': 4, 'word_size': 4, 'window_step': 0.2,
       'numerosity_reduction': False},
      X_bow[:2], ['abcd abcd abcd abcd', 'abcd abcd abcd abcd']),

     ({'window_size': 4, 'word_size': 4, 'window_step': 0.5},
      X_bow[:2], ['abcd', 'abcd'])]
)
def test_actual_results_bag_of_words(params, X, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = BagOfWords(**params).fit_transform(X)
    np.testing.assert_array_equal(arr_actual, arr_desired)



"""Testing for Singular Spectrum Analysis."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
import re
from pyts.decomposition.ssa import _outer_dot, _diagonal_averaging
from pyts.decomposition import SingularSpectrumAnalysis


rng = np.random.RandomState(42)
X = rng.randn(4, 30)


@pytest.mark.parametrize(
    'v, X, arr_desired',
    [(np.zeros((2, 2, 2)), np.ones((2, 2, 2)), np.zeros((2, 2, 2, 2))),

     (np.ones((2, 2, 2)), np.zeros((2, 2, 2)), np.zeros((2, 2, 2, 2))),

     ([[[1, 2], [3, 4]]],
      np.ones((1, 2, 2)),
      [[[[4, 4], [12, 12]], [[12, 12], [24, 24]]]]),

     ([[[1, 2], [3, 4]]],
      [[[2, 1], [-2, -3]]],
      [[[[-4, -8], [-12, -24]], [[-8, -20], [-16, -40]]]])]
)
def test_outer_dot(v, X, arr_desired):
    """Test that the actual results are the expected ones."""
    v = np.asarray(v).astype('float64')
    X = np.asarray(X).astype('float64')
    n_samples, n_windows, window_size = X.shape
    arr_actual = _outer_dot(v, X, n_samples, window_size, n_windows)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'X, arr_desired',
    [([[[[0, 1, 2], [3, 4, 5]]]], [[[0, (1 + 3) / 2, (2 + 4) / 2, 5]]]),

     ([[[[0, 1, 2, 0, -1], [3, -2, 4, 5, -3], [1, 3, 2, 4, 5]]]],
      [[[0, (1 + 3) / 2, (2 - 2 + 1) / 3, (0 + 4 + 3) / 3,
         (-1 + 5 + 2) / 3, (-3 + 4) / 2, 5]]])]
)
def test_diagonal_averaging(X, arr_desired):
    """Test that the actual results are the expected ones."""
    X = np.asarray(X, dtype='float64')
    n_samples, grouping_size, window_size, n_windows = X.shape
    n_timestamps = window_size + n_windows - 1
    gap = max(n_windows, window_size)

    arr_actual = _diagonal_averaging(
        X, n_samples, n_timestamps, window_size,
        n_windows, grouping_size, gap
    )
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'window_size': '4'}, TypeError,
      "'window_size' must be an integer or a float."),

     ({'groups': '3'}, TypeError,
      "'groups' must be either None, an integer or array-like."),

     ({'window_size': 1}, ValueError,
      "If 'window_size' is an integer, it must be greater than or equal to 2 "
      "and lower than or equal to n_timestamps (got 1)."),

     ({'window_size': 0.}, ValueError,
      "If 'window_size' is a float, it must be greater than 0 and lower than "
      "or equal to 1 (got {0}).".format(0.)),

     ({'groups': 6}, ValueError,
      "If 'groups' is an integer, it must be greater than or equal to 1 "
      "and lower than or equal to 'window_size'."),

     ({'groups': [[0, 2, 5]]}, ValueError,
      "If 'groups' is array-like, all the values in 'groups' must be integers "
      "between 0 and ('window_size' - 1).")]
)
def test_parameter_check(params, error, err_msg):
    """Test parameter validation."""
    ssa = SingularSpectrumAnalysis(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        ssa.transform(X)


@pytest.mark.parametrize(
    'params',
    [({}),
     ({'window_size': 2}),
     ({'window_size': 25}),
     ({'window_size': 0.2}),
     ({'groups': 3}),
     ({'window_size': 10, 'groups': 10}),
     ({'groups': [[0, 1], [2, 3]]}),
     ({'groups': [[0, 2], [1, 3]]}),
     ({'window_size': 8, 'groups': [[0, 2, 4, 6], [1, 3, 5, 7]]})]
)
def test_actual_results(params):
    """Test that the actual results are the expected ones."""
    ssa = SingularSpectrumAnalysis(**params)
    arr_actual = ssa.fit_transform(X).sum(axis=1)
    np.testing.assert_allclose(arr_actual, X, atol=1e-5, rtol=0.)



"""Code for k-nearest-neighbors."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

from sklearn.base import BaseEstimator
from sklearn.neighbors import KNeighborsClassifier as SklearnKNN
from sklearn.preprocessing import LabelEncoder
from sklearn.utils.validation import check_X_y, check_is_fitted
from ..base import UnivariateClassifierMixin
from ..metrics import boss, dtw, sakoe_chiba_band, itakura_parallelogram
from ..metrics.dtw import (_dtw_classic, _dtw_region, _dtw_fast,
                           _dtw_multiscale)


class KNeighborsClassifier(BaseEstimator, UnivariateClassifierMixin):
    """k-nearest neighbors classifier.

    Parameters
    ----------
    n_neighbors : int, optional (default = 1)
        Number of neighbors to use.

    weights : str or callable, optional (default = 'uniform')
        weight function used in prediction.  Possible values:

        - 'uniform' : uniform weights.  All points in each neighborhood
          are weighted equally.
        - 'distance' : weight points by the inverse of their distance.
          in this case, closer neighbors of a query point will have a
          greater influence than neighbors which are further away.
        - [callable] : a user-defined function which accepts an
          array of distances, and returns an array of the same shape
          containing the weights.

    algorithm : {'auto', 'ball_tree', 'kd_tree', 'brute'}, optional
        Algorithm used to compute the nearest neighbors. Ignored ff ``metric``
        is either 'dtw', 'dtw_sakoechiba', 'dtw_itakura', 'dtw_multiscale',
        'dtw_fast' or 'boss' ('brute' will be used).

        Note: fitting on sparse input will override the setting of
        this parameter, using brute force.

    leaf_size : int, optional (default = 30)
        Leaf size passed to BallTree or KDTree.  This can affect the
        speed of the construction and query, as well as the memory
        required to store the tree.  The optimal value depends on the
        nature of the problem.

    metric : string or DistanceMetric object (default = 'minkowski')
        The distance metric to use for the tree.  The default metric is
        minkowski, and with p=2 is equivalent to the standard Euclidean
        metric. See the documentation of the DistanceMetric class from
        scikit-learn for a list of available metrics.
        For Dynamic Time Warping, the available metrics are 'dtw',
        'dtw_sakoechiba', 'dtw_itakura', 'dtw_multiscale' and 'dtw_fast'.
        For BOSS metric, one can use 'boss'.

    p : integer, optional (default = 2)
        Power parameter for the Minkowski metric. When p = 1, this is
        equivalent to using manhattan_distance (l1), and euclidean_distance
        (l2) for p = 2. For arbitrary p, minkowski_distance (l_p) is used.

    metric_params : dict, optional (default = None)
        Additional keyword arguments for the metric function.

    n_jobs : int, optional (default = 1)
        The number of parallel jobs to run for neighbors search.
        If ``n_jobs=-1``, then the number of jobs is set to the number of CPU
        cores. Doesn't affect :meth:`fit` method.

    Attributes
    ----------
    classes_ : array, shape = (n_classes,)
        An array of class labels known to the classifier.

    Examples
    --------
    >>> from pyts.classification import KNeighborsClassifier
    >>> from pyts.datasets import load_gunpoint
    >>> X_train, X_test, y_train, y_test = load_gunpoint(return_X_y=True)
    >>> clf = KNeighborsClassifier()
    >>> clf.fit(X_train, y_train)
    KNeighborsClassifier(...)
    >>> clf.score(X_test, y_test)
    0.91...

    """

    def __init__(self, n_neighbors=1, weights='uniform', algorithm='auto',
                 leaf_size=30, p=2, metric='minkowski', metric_params=None,
                 n_jobs=1, **kwargs):
        self.n_neighbors = n_neighbors
        self.weights = weights
        self.algorithm = algorithm
        self.leaf_size = leaf_size
        self.p = p
        self.metric = metric
        self.metric_params = metric_params
        self.n_jobs = n_jobs
        self.kwargs = kwargs

    def fit(self, X, y):
        """Fit the model according to the given training data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Training vector.

        y : array-like, shape = (n_samples,)
            Class labels for each data sample.

        Returns
        -------
        self : object

        """
        X, y = check_X_y(X, y)
        self._le = LabelEncoder().fit(y)
        self.classes_ = self._le.classes_

        if self.metric == 'dtw':
            self._clf = SklearnKNN(
                n_neighbors=self.n_neighbors, weights=self.weights,
                algorithm='brute', metric=dtw,
                metric_params=self.metric_params,
                n_jobs=self.n_jobs, **self.kwargs
            )

        elif self.metric == 'dtw_classic':
            self._clf = SklearnKNN(
                n_neighbors=self.n_neighbors, weights=self.weights,
                algorithm='brute', metric=_dtw_classic,
                metric_params=self.metric_params,
                n_jobs=self.n_jobs, **self.kwargs
            )

        elif self.metric == 'dtw_sakoechiba':
            n_timestamps = X.shape[1]
            if self.metric_params is None:
                region = sakoe_chiba_band(n_timestamps)
            else:
                if 'window_size' not in self.metric_params.keys():
                    window_size = 0.1
                else:
                    window_size = self.metric_params['window_size']
                region = sakoe_chiba_band(n_timestamps,
                                          window_size=window_size)
            self._clf = SklearnKNN(
                n_neighbors=self.n_neighbors, weights=self.weights,
                algorithm='brute', metric=_dtw_region,
                metric_params={'region': region},
                n_jobs=self.n_jobs, **self.kwargs
            )

        elif self.metric == 'dtw_itakura':
            n_timestamps = X.shape[1]
            if self.metric_params is None:
                region = itakura_parallelogram(n_timestamps)
            else:
                if 'max_slope' not in self.metric_params.keys():
                    max_slope = 2.
                else:
                    max_slope = self.metric_params['max_slope']
                region = itakura_parallelogram(n_timestamps,
                                               max_slope=max_slope)
            self._clf = SklearnKNN(
                n_neighbors=self.n_neighbors, weights=self.weights,
                algorithm='brute', metric=_dtw_region,
                metric_params={'region': region},
                n_jobs=self.n_jobs, **self.kwargs
            )

        elif self.metric == 'dtw_multiscale':
            self._clf = SklearnKNN(
                n_neighbors=self.n_neighbors, weights=self.weights,
                algorithm='brute', metric=_dtw_multiscale,
                metric_params=self.metric_params,
                n_jobs=self.n_jobs, **self.kwargs
            )

        elif self.metric == 'dtw_fast':
            self._clf = SklearnKNN(
                n_neighbors=self.n_neighbors, weights=self.weights,
                algorithm='brute', metric=_dtw_fast,
                metric_params=self.metric_params,
                n_jobs=self.n_jobs, **self.kwargs
            )

        elif self.metric == 'boss':
            self._clf = SklearnKNN(
                n_neighbors=self.n_neighbors, weights=self.weights,
                algorithm='brute', metric=boss,
                metric_params=self.metric_params,
                n_jobs=self.n_jobs, **self.kwargs
            )

        else:
            self._clf = SklearnKNN(
                n_neighbors=self.n_neighbors, weights=self.weights,
                algorithm=self.algorithm, leaf_size=self.leaf_size,
                p=self.p, metric=self.metric, metric_params=self.metric_params,
                n_jobs=self.n_jobs, **self.kwargs
            )

        self._clf.fit(X, y)
        return self

    def predict_proba(self, X):
        """Return probability estimates for the test data X.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Test samples.

        Returns
        -------
        p : array, shape = (n_samples, n_classes)
            Probability estimates.

        """
        check_is_fitted(self, '_clf')
        return self._clf.predict_proba(X)

    def predict(self, X):
        """Predict the class labels for the provided data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Test samples.

        Returns
        -------
        y_pred : array-like, shape = (n_samples,)
            Class labels for each data sample.

        """
        check_is_fitted(self, '_clf')
        return self._clf.predict(X)



"""Testing for Bag-of-SFA Symbols in Vector Space."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
import re
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.feature_extraction.text import TfidfVectorizer
from pyts.classification import BOSSVS
from pyts.approximation import SymbolicFourierApproximation


rng = np.random.RandomState(42)
X = rng.randn(8, 20)
y = np.asarray([0, 0, 0, 0, 1, 1, 1, 1])


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'word_size': '4'}, TypeError, "'word_size' must be an integer."),

     ({'window_size': None}, TypeError,
      "'window_size' must be an integer or a float."),

     ({'window_step': '2'}, TypeError,
      "'window_step' must be an integer or a float."),

     ({'word_size': 0}, ValueError, "'word_size' must be a positive integer."),

     ({'window_size': 0}, ValueError,
      "If 'window_size' is an integer, it must be greater than or equal to 1 "
      "and lower than or equal to n_timestamps if 'drop_sum=False'."),

     ({'window_size': 0, 'drop_sum': True}, ValueError,
      "If 'window_size' is an integer, it must be greater than or equal to 1 "
      "and lower than or equal to (n_timestamps - 1) if 'drop_sum=True'."),

     ({'window_size': 2.}, ValueError,
      "If 'window_size' is a float, it must be greater than 0 and lower than "
      "or equal to 1."),

     ({'window_step': 0}, ValueError,
      "If 'window_step' is an integer, it must be greater than or equal to 1 "
      "and lower than or equal to n_timestamps."),

     ({'window_step': 2.}, ValueError,
      "If 'window_step' is a float, it must be greater than 0 and lower than "
      "or equal to 1."),

     ({'window_size': 4, 'word_size': 4, 'drop_sum': True}, ValueError,
      "'word_size' must be lower than or equal to (window_size - 1) if "
      "'drop_sum=True'."),

     ({'window_size': 4, 'word_size': 5}, ValueError,
      "'word_size' must be lower than or equal to window_size if "
      "'drop_sum=False'.")]
)
def test_parameter_check(params, error, err_msg):
    """Test parameter validation."""
    clf = BOSSVS(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        clf.fit(X, y)


def test_actual_results_no_numerosity_reduction():
    """Test that the actual results are the expected ones."""
    bossvs = BOSSVS(
        word_size=4, n_bins=3, window_size=10, window_step=10,
        anova=False, drop_sum=False, norm_mean=False, norm_std=False,
        strategy='quantile', alphabet=None, numerosity_reduction=False,
        use_idf=True, smooth_idf=False, sublinear_tf=True
    )

    X_windowed = X.reshape(8, 2, 10).reshape(16, 10)
    sfa = SymbolicFourierApproximation(
        n_coefs=4, drop_sum=False, anova=False, norm_mean=False,
        norm_std=False, n_bins=3, strategy='quantile', alphabet=None
    )
    y_repeated = np.repeat(y, 2)
    X_sfa = sfa.fit_transform(X_windowed, y_repeated)
    X_word = np.asarray([''.join(X_sfa[i])
                         for i in range(16)])
    X_word = X_word.reshape(8, 2)
    X_bow = np.asarray([' '.join(X_word[i]) for i in range(8)])
    X_class = np.array([' '.join(X_bow[y == i]) for i in range(2)])

    tfidf = TfidfVectorizer(
        norm=None, use_idf=True, smooth_idf=False, sublinear_tf=True
    )
    tfidf_desired = tfidf.fit_transform(X_class).toarray()

    # Vocabulary
    vocabulary_desired = {value: key for key, value in
                          tfidf.vocabulary_.items()}

    # Tf-idf
    tfidf_actual = bossvs.fit(X, y).tfidf_

    # Decision function
    decision_function_actual = bossvs.decision_function(X)
    decision_function_desired = cosine_similarity(
        tfidf.transform(X_bow), tfidf_desired)

    # Predictions
    y_pred_actual = bossvs.predict(X)
    y_pred_desired = decision_function_desired.argmax(axis=1)

    # Testing
    assert bossvs.vocabulary_ == vocabulary_desired
    np.testing.assert_allclose(tfidf_actual, tfidf_desired, atol=1e-5, rtol=0)
    np.testing.assert_allclose(
        decision_function_actual, decision_function_desired, atol=1e-5, rtol=0)
    np.testing.assert_allclose(
        y_pred_actual, y_pred_desired, atol=1e-5, rtol=0)


def test_actual_results_numerosity_reduction():
    """Test that the actual results are the expected ones."""
    bossvs = BOSSVS(
        word_size=4, n_bins=3, window_size=0.5, window_step=0.5,
        anova=False, drop_sum=False, norm_mean=False, norm_std=False,
        strategy='quantile', alphabet=None, numerosity_reduction=True,
        use_idf=False, smooth_idf=False, sublinear_tf=True
    )

    X_windowed = X.reshape(8, 2, 10).reshape(16, 10)
    sfa = SymbolicFourierApproximation(
        n_coefs=4, drop_sum=False, anova=False, norm_mean=False,
        norm_std=False, n_bins=3, strategy='quantile', alphabet=None
    )
    y_repeated = np.repeat(y, 2)
    X_sfa = sfa.fit_transform(X_windowed, y_repeated)
    X_word = np.asarray([''.join(X_sfa[i])
                         for i in range(16)])
    X_word = X_word.reshape(8, 2)
    not_equal = np.c_[X_word[:, 1:] != X_word[:, :-1], np.full(8, True)]
    X_bow = np.asarray([' '.join(X_word[i, not_equal[i]]) for i in range(8)])
    X_class = np.array([' '.join(X_bow[y == i]) for i in range(2)])

    tfidf = TfidfVectorizer(
        norm=None, use_idf=False, smooth_idf=False, sublinear_tf=True
    )
    tfidf_desired = tfidf.fit_transform(X_class).toarray()
    vocabulary_desired = {value: key for key, value in
                          tfidf.vocabulary_.items()}

    # Tf-idf
    tfidf_actual = bossvs.fit(X, y).tfidf_

    # Decision function
    decision_function_actual = bossvs.decision_function(X)
    decision_function_desired = cosine_similarity(
        tfidf.transform(X_bow), tfidf_desired)

    # Predictions
    y_pred_actual = bossvs.predict(X)
    y_pred_desired = decision_function_desired.argmax(axis=1)

    # Testing
    assert bossvs.vocabulary_ == vocabulary_desired
    np.testing.assert_allclose(tfidf_actual, tfidf_desired, atol=1e-5, rtol=0)
    np.testing.assert_allclose(
        decision_function_actual, decision_function_desired, atol=1e-5, rtol=0)
    np.testing.assert_allclose(
        y_pred_actual, y_pred_desired, atol=1e-5, rtol=0)



"""Testing for Markov Transition Field."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
import re
from pyts.image.mtf import (_markov_transition_matrix,
                            _markov_transition_field,
                            _aggregated_markov_transition_field)
from pyts.image import MarkovTransitionField


X = [[0, 1, 2, 3], [1, 0, 1, 1]]


@pytest.mark.parametrize(
    'X_binned, n_bins, arr_desired',
    [([[0, 1, 1, 0, 0, 1, 0, 0, 1, 0, 0]], 2, [[[3, 3], [3, 1]]]),

     ([[0, 1, 1, 0, 0, 1, 0, 0, 1, 0, 0]], 3,
      [[[3, 3, 0], [3, 1, 0], [0, 0, 0]]]),

     ([[0, 1, 2, 3], [0, 2, 1, 3]], 4,
      [[[0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1], [0, 0, 0, 0]],
       [[0, 0, 1, 0], [0, 0, 0, 1], [0, 1, 0, 0], [0, 0, 0, 0]]])]
)
def test_markov_transition_matrix(X_binned, n_bins, arr_desired):
    """Test that the actual results are the expected ones."""
    X_binned = np.asarray(X_binned)
    n_samples, n_timestamps = X_binned.shape
    arr_actual = _markov_transition_matrix(
        X_binned, n_samples, n_timestamps, n_bins)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'X_binned, X_mtm, n_bins, arr_desired',
    [([[1, 0, 1, 0, 0, 1, 0]], [[[1, 2], [3, 0]]], 2,
      [[[0, 3, 0, 3, 3, 0, 3],
        [2, 1, 2, 1, 1, 2, 1],
        [0, 3, 0, 3, 3, 0, 3],
        [2, 1, 2, 1, 1, 2, 1],
        [2, 1, 2, 1, 1, 2, 1],
        [0, 3, 0, 3, 3, 0, 3],
        [2, 1, 2, 1, 1, 2, 1]]]),

     ([[0, 1, 2, 0, 1], [0, 2, 0, 1, 1]],
      [[[0, 1, 0], [0, 0, 1], [1, 0, 0]], [[0, .5, .5], [0, 1, 0], [1, 0, 0]]],
      3,
      [[[0, 1, 0, 0, 1],
        [0, 0, 1, 0, 0],
        [1, 0, 0, 1, 0],
        [0, 1, 0, 0, 1],
        [0, 0, 1, 0, 0]],
       [[0.0, 0.5, 0.0, 0.5, 0.5],
        [1.0, 0.0, 1.0, 0.0, 0.0],
        [0.0, 0.5, 0.0, 0.5, 0.5],
        [0.0, 0.0, 0.0, 1.0, 1.0],
        [0.0, 0.0, 0.0, 1.0, 1.0]]])]
)
def test_markov_transition_field(X_binned, X_mtm, n_bins, arr_desired):
    """Test that the actual results are the expected ones."""
    X_binned = np.asarray(X_binned)
    X_mtm = np.asarray(X_mtm)
    n_samples, n_timestamps = X_binned.shape
    arr_actual = _markov_transition_field(
        X_binned, X_mtm, n_samples, n_timestamps, n_bins
    )
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'image_size, start, end, X_mtf, arr_desired',
    [(2, [0, 2], [2, 4],
      [[[0, 1, 2, 0], [1, 0, 2, 0], [1, 1, 0, 0], [0, 1, 2, 2]],
       [[2, 1, 2, 0], [0, 1, 3, 0], [0, 1, 2, 0], [0, 0, 0, 0]]],
      [[[0.5, 1.0], [0.75, 1.0]], [[1.0, 1.25], [0.25, 0.5]]]),

     (3, [0, 1, 2], [2, 3, 4],
      [[[0, 1, 2, 0], [1, 0, 2, 0], [1, 1, 0, 0], [0, 1, 2, 2]],
       [[2, 1, 2, 0], [0, 1, 3, 0], [0, 1, 2, 0], [0, 0, 0, 0]]],
      [[[0.5, 1.25, 1.0], [0.75, 0.75, 0.5], [0.75, 1, 1.0]],
       [[1.0, 1.75, 1.25], [0.5, 1.75, 1.25], [0.25, 0.75, 0.5]]])]
)
def test_aggregated_markov_transition_field(
    image_size, start, end, X_mtf, arr_desired
):
    """Test that the actual results are the expected ones."""
    X_mtf = np.asarray(X_mtf)
    start = np.asarray(start)
    end = np.asarray(end)
    n_samples, _ = X_mtf.shape[:2]
    arr_actual = _aggregated_markov_transition_field(
        X_mtf, n_samples, image_size, start, end)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'image_size': '4'}, TypeError,
      "'image_size' must be an integer or a float."),

     ({'n_bins': [0, 1]}, TypeError,
      "'n_bins' must be an integer."),

     ({'image_size': 0}, ValueError,
      "If 'image_size' is an integer, it must be greater than or equal to 1 "
      "and lower than or equal to n_timestamps (got 0)."),

     ({'image_size': 2.}, ValueError,
      "If 'image_size' is a float, it must be greater than 0 and lower than "
      "or equal to 1 (got {0}).".format(2.)),

     ({'n_bins': 1}, ValueError,
      "'n_bins' must be greater than or equal to 2."),

     ({'strategy': 'whoops'}, ValueError,
      "'strategy' must be 'uniform', 'quantile' or 'normal'.")]
)
def test_parameter_check(params, error, err_msg):
    """Test parameter validation."""
    mtf = MarkovTransitionField(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        mtf.transform(X)


@pytest.mark.parametrize(
    'params, X, arr_desired',
    [({'image_size': 2, 'n_bins': 4},
      [np.arange(8)], [[[0.375, 0.125], [0.000, 0.500]]]),

     ({'image_size': 2, 'n_bins': 4, 'strategy': 'uniform'},
      [np.arange(8)], [[[0.375, 0.125], [0.000, 0.500]]]),

     ({'image_size': 0.25, 'n_bins': 4},
      [np.arange(8)], [[[0.375, 0.125], [0.000, 0.500]]]),

     ({'image_size': 7, 'n_bins': 2, 'overlapping': True},
      [np.arange(8)],
      [[[0.75, 0.75, 0.75, 0.50, 0.25, 0.25, 0.25],
        [0.75, 0.75, 0.75, 0.50, 0.25, 0.25, 0.25],
        [0.75, 0.75, 0.75, 0.50, 0.25, 0.25, 0.25],
        [0.375, 0.375, 0.375, 0.5, 0.625, 0.625, 0.625],
        [0, 0, 0, 0.5, 1, 1, 1],
        [0, 0, 0, 0.5, 1, 1, 1],
        [0, 0, 0, 0.5, 1, 1, 1]]]),

     ({'image_size': 7, 'n_bins': 2, 'overlapping': True,
       'strategy': 'uniform'}, [np.arange(8)],
      [[[0.75, 0.75, 0.75, 0.50, 0.25, 0.25, 0.25],
        [0.75, 0.75, 0.75, 0.50, 0.25, 0.25, 0.25],
        [0.75, 0.75, 0.75, 0.50, 0.25, 0.25, 0.25],
        [0.375, 0.375, 0.375, 0.5, 0.625, 0.625, 0.625],
        [0, 0, 0, 0.5, 1, 1, 1],
        [0, 0, 0, 0.5, 1, 1, 1],
        [0, 0, 0, 0.5, 1, 1, 1]]]),

     ({'image_size': 2, 'n_bins': 2, 'strategy': 'uniform'}, X,
      [[[0.5, 0.5], [0, 1]], [[0.5, 0.75], [0.5, 0.5]]])]
)
def test_actual_results(params, X, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = MarkovTransitionField(**params).fit_transform(X)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


def test_flatten():
    """Test the 'flatten' parameter."""
    arr_false = MarkovTransitionField(n_bins=2).transform(X).reshape(2, -1)
    arr_true = MarkovTransitionField(n_bins=2, flatten=True).transform(X)
    np.testing.assert_allclose(arr_false, arr_true, atol=1e-5, rtol=0.)



"""Testing for Discrete Fourier Transform."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
import re
from sklearn.feature_selection import f_classif
from pyts.approximation import DiscreteFourierTransform


rng = np.random.RandomState(42)
n_samples, n_timestamps = 5, 8
X_even = rng.randn(n_samples, n_timestamps)
X_odd = X_even[:, :-1]
y = rng.randint(2, size=n_samples)


def _compute_expected_results(X, y=None, n_coefs=None, drop_sum=False,
                              anova=False, norm_mean=False, norm_std=False):
    """Compute the expected results."""
    X = np.asarray(X)
    n_samples, n_timestamps = X.shape
    if norm_mean:
        X -= X.mean(axis=1)[:, None]
    if norm_std:
        X /= X.std(axis=1)[:, None]
    X_fft = np.fft.rfft(X)
    X_fft = np.vstack([np.real(X_fft), np.imag(X_fft)])
    if n_timestamps % 2 == 0:
        X_fft = X_fft.reshape(n_samples, n_timestamps + 2, order='F')
        X_fft = np.c_[X_fft[:, 0], X_fft[:, 2:-1]]
    else:
        X_fft = X_fft.reshape(n_samples, n_timestamps + 1, order='F')
        X_fft = np.c_[X_fft[:, 0], X_fft[:, 2:]]
    if drop_sum:
        X_fft = X_fft[:, 1:]
    if n_coefs is None:
        return X_fft
    else:
        if anova:
            _, p = f_classif(X_fft, y)
            support = np.argsort(p)[:n_coefs]
            return X_fft[:, support]
        else:
            return X_fft[:, :n_coefs]


@pytest.mark.parametrize(
    'params, X, error, err_msg',
    [({'n_coefs': '3'}, X_even, TypeError,
      "'n_coefs' must be None, an integer or a float."),

     ({'n_coefs': 0}, X_even, ValueError,
      "If 'n_coefs' is an integer, it must be greater than or equal to 1 and "
      "lower than or equal to n_timestamps if 'drop_sum=False'."),

     ({'n_coefs': 8, 'drop_sum': True}, X_even, ValueError,
      "If 'n_coefs' is an integer, it must be greater than or equal to 1 and "
      "lower than or equal to (n_timestamps - 1) if 'drop_sum=True'."),

     ({'n_coefs': 2.}, X_even, ValueError,
      ("If 'n_coefs' is a float, it must be greater than 0 and lower than or "
       "equal to 1.")),

     ({'anova': True, 'n_coefs': 2}, np.zeros_like(X_even), ValueError,
      ("All the Fourier coefficients are constant. Your input data is weirdly "
       "homogeneous.")),

     ({'anova': True, 'n_coefs': 2, 'drop_sum': True},
      np.arange(100).reshape(5, 20), ValueError,
      ("All the Fourier coefficients are constant. Your input data is weirdly "
       "homogeneous."))]
)
def test_parameter_check(params, X, error, err_msg):
    """Test parameter validation."""
    dft = DiscreteFourierTransform(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        dft.fit(X, y)


@pytest.mark.parametrize(
    'params, X, n_non_zeros',
    [({'n_coefs': 3}, np.arange(100).reshape(5, 20), 1),
     ({'n_coefs': 4}, np.arange(100).reshape(5, 20), 1),
     ({'n_coefs': 5}, np.arange(100).reshape(5, 20), 1)]
)
def test_parameter_check_anova_warning(params, X, n_non_zeros):
    dft = DiscreteFourierTransform(**params, anova=True)
    msg = (
        "The number of non constant Fourier coefficients \\({0}\\) "
        "is lower than the number of coefficients to keep \\({1}\\). "
        "The number of coefficients to keep is truncated to {2}."
    ).format(n_non_zeros, params['n_coefs'], n_non_zeros)
    with pytest.warns(UserWarning, match=msg):
        dft.fit(X, np.random.randint(2, size=X.shape[0]))


@pytest.mark.parametrize('X', [X_even, X_odd])
@pytest.mark.parametrize(
    'params',
    [({}),
     ({'n_coefs': 3}),
     ({'drop_sum': True}),
     ({'anova': True}),
     ({'norm_mean': True}),
     ({'norm_std': True}),
     ({'norm_mean': True, 'norm_std': True}),
     ({'n_coefs': 2, 'drop_sum': True, 'anova': True})]
)
def test_actual_results(X, params):
    """Test that the actual results are the expected ones."""
    arr_actual = DiscreteFourierTransform(**params).fit_transform(X, y)
    arr_desired = _compute_expected_results(X, y, **params)
    np.testing.assert_array_equal(arr_actual, arr_desired)


@pytest.mark.parametrize('X', [X_even, X_odd])
@pytest.mark.parametrize(
    'params',
    [({}),
     ({'n_coefs': 3}),
     ({'n_coefs': 0.5}),
     ({'n_coefs': 0.5, 'drop_sum': True}),
     ({'drop_sum': True}),
     ({'anova': True}),
     ({'norm_mean': True}),
     ({'norm_std': True}),
     ({'norm_mean': True, 'norm_std': True}),
     ({'n_coefs': 2, 'drop_sum': True, 'anova': True})]
)
def test_fit_transform(X, params):
    """Test that fit and transform yield the same results as fit_transform."""
    arr_1 = DiscreteFourierTransform(**params).fit(X, y).transform(X)
    arr_2 = DiscreteFourierTransform(**params).fit_transform(X, y)
    np.testing.assert_array_equal(arr_1, arr_2)



"""Code for Bag-of-patterns representation for time series."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

from scipy.sparse import csr_matrix
from sklearn.base import BaseEstimator
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.utils.validation import check_array, check_is_fitted
from ..bag_of_words import BagOfWords
from ..base import UnivariateTransformerMixin


class BagOfPatterns(BaseEstimator, UnivariateTransformerMixin):
    """Bag-of-patterns representation for time series.

    This algorithm uses a sliding window to extract subsequences from the
    time series and transforms each subsequence into a word using the
    Piecewise Aggregate Approximation and the Symbolic Aggregate approXimation
    algorithms. Thus it transforms each time series into a bag of words.
    Then it derives the frequencies of each word for each time series.

    Parameters
    ----------
    window_size : int or float (default = 0.5)
        Length of the sliding window. If float, it represents
        a percentage of the size of each time series and must be
        between 0 and 1.

    word_size : int or float (default = 0.5)
        Length of the words. If float, it represents
        a percentage of the length of the sliding window and must be
        between 0. and 1.

    n_bins : int (default = 4)
        The number of bins to produce. It must be between 2 and
        ``min(window_size, 26)``.

    strategy : 'uniform', 'quantile' or 'normal' (default = 'normal')
        Strategy used to define the widths of the bins:

        - 'uniform': All bins in each sample have identical widths
        - 'quantile': All bins in each sample have the same number of points
        - 'normal': Bin edges are quantiles from a standard normal distribution

    numerosity_reduction : bool (default = True)
        If True, delete sample-wise all but one occurence of back to back
        identical occurences of the same words.

    window_step : int or float (default = 1)
        Step of the sliding window. If float, it represents the percentage of
        the size of each time series and must be between 0 and 1. The step of
        sliding window will be computed as
        ``ceil(window_step * n_timestamps)``.

    norm_mean : bool (default = True)
        If True, center each subseries before scaling.

    norm_std : bool (default = True)
        If True, scale each subseries to unit variance.

    sparse : bool (default = True)
        Return a sparse matrix if True, else return an array.

    overlapping : bool (default = True)
        If True, time points may belong to two bins when decreasing the size
        of the subsequence with the Piecewise Aggregate Approximation
        algorithm. If False, each time point belong to one single bin, but
        the size of the bins may vary.

    alphabet : None or array-like, shape = (n_bins,)
        Alphabet to use. If None, the first `n_bins` letters of the Latin
        alphabet are used.

    Attributes
    ----------
    vocabulary_ : dict
        A mapping of feature indices to terms.

    References
    ----------
    .. [1] J. Lin, R. Khade and Y. Li, "Rotation-invariant similarity in time
           series using bag-of-patterns representation". Journal of Intelligent
           Information Systems, 39 (2), 287-315 (2012).

    Examples
    --------
    >>> import numpy as np
    >>> from pyts.transformation import BagOfPatterns
    >>> X = np.arange(12).reshape(2, 6)
    >>> bop = BagOfPatterns(window_size=4, word_size=4, sparse=False)
    >>> bop.fit_transform(X)
    array(...)
    >>> bop.set_params(numerosity_reduction=False)
    BagOfPatterns(...)
    >>> bop.fit_transform(X)
    array(...)

    """

    def __init__(self, window_size=0.5, word_size=0.5, n_bins=4,
                 strategy='normal', numerosity_reduction=True, window_step=1,
                 norm_mean=True, norm_std=True, sparse=True, overlapping=True,
                 alphabet=None):
        self.window_size = window_size
        self.word_size = word_size
        self.n_bins = n_bins
        self.strategy = strategy
        self.numerosity_reduction = numerosity_reduction
        self.window_step = window_step
        self.norm_mean = norm_mean
        self.norm_std = norm_std
        self.sparse = sparse
        self.overlapping = overlapping
        self.alphabet = alphabet

    def fit(self, X, y=None):
        """Learn the dictionary.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Input data

        y
            Ignored

        Returns
        -------
        self : object

        """
        # Transform each time series into a bag of words
        bow = BagOfWords(
            window_size=self.window_size, word_size=self.word_size,
            n_bins=self.n_bins, strategy=self.strategy,
            numerosity_reduction=self.numerosity_reduction,
            window_step=self.window_step, norm_mean=self.norm_mean,
            norm_std=self.norm_std, overlapping=self.overlapping,
            alphabet=self.alphabet
        )
        X_bow = bow.transform(X)

        # Learn the vocabulary
        vectorizer = CountVectorizer()
        vectorizer.fit(X_bow)
        self.vocabulary_ = {value: key for key, value in
                            vectorizer.vocabulary_.items()}
        self._vectorizer = vectorizer
        return self

    def transform(self, X):
        """Derive word frequencies for each time series.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Data to transform.

        Returns
        -------
        X_new : array, shape = (n_samples, n_words)
            Word frequencies.

        """
        X = check_array(X, dtype='float64')
        check_is_fitted(self, 'vocabulary_')

        # Transform each time series into a bag of words
        bow = BagOfWords(
            window_size=self.window_size, word_size=self.word_size,
            n_bins=self.n_bins, strategy=self.strategy,
            numerosity_reduction=self.numerosity_reduction,
            window_step=self.window_step, norm_mean=self.norm_mean,
            norm_std=self.norm_std, overlapping=self.overlapping,
            alphabet=self.alphabet
        )
        X_bow = bow.transform(X)

        # Derive frequencies for each word in the vocabulary
        X_bop = self._vectorizer.transform(X_bow)
        if not self.sparse:
            return X_bop.A
        return csr_matrix(X_bop)

    def fit_transform(self, X, y=None):
        """Derive word frequencies for each time series.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Data to transform.

        y
            Ignored

        Returns
        -------
        X_new : array, shape = (n_samples, n_words)
            Word frequencies.

        """
        # Transform each time series into a bag of words
        bow = BagOfWords(
            window_size=self.window_size, word_size=self.word_size,
            n_bins=self.n_bins, strategy=self.strategy,
            numerosity_reduction=self.numerosity_reduction,
            window_step=self.window_step, norm_mean=self.norm_mean,
            norm_std=self.norm_std, overlapping=self.overlapping,
            alphabet=self.alphabet
        )
        X_bow = bow.transform(X)

        # Derive frequencies of each word
        vectorizer = CountVectorizer()
        X_bop = vectorizer.fit_transform(X_bow)
        self.vocabulary_ = {value: key for key, value in
                            vectorizer.vocabulary_.items()}
        self._vectorizer = vectorizer
        if not self.sparse:
            return X_bop.A
        return csr_matrix(X_bop)



"""Code for Bag-of-words representation for time series."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
from math import ceil
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.pipeline import make_pipeline
from sklearn.utils.validation import check_array
from ..approximation import (
    PiecewiseAggregateApproximation, SymbolicAggregateApproximation)
from ..base import UnivariateTransformerMixin
from ..preprocessing import KBinsDiscretizer, StandardScaler
from ..preprocessing.discretizer import _digitize
from ..utils.utils import _windowed_view


class WordExtractor(BaseEstimator, UnivariateTransformerMixin):
    r"""Transform discretized time series into sequences of words.

    Parameters
    ----------
    window_size : int or float (default = 0.1)
        Size of the sliding window (i.e. the size of each word). If float, it
        represents the percentage of the size of each time series and must be
        between 0 and 1. The window size will be computed as
        ``ceil(window_size * n_timestamps)``.

    window_step : int or float (default = 1)
        Step of the sliding window. If float, it represents the percentage of
        the size of each time series and must be between 0 and 1. The window
        size will be computed as ``ceil(window_step * n_timestamps)``.

    numerosity_reduction : bool (default = True)
        If True, delete sample-wise all but one occurence of back to back
        identical occurences of the same words.

    Examples
    --------
    >>> from pyts.bag_of_words import WordExtractor
    >>> X = [['a', 'a', 'b', 'a', 'b', 'b', 'b', 'b', 'a'],
    ...      ['a', 'b', 'c', 'c', 'c', 'c', 'a', 'a', 'c']]
    >>> word = WordExtractor(window_size=2)
    >>> print(word.transform(X))
    ['aa ab ba ab bb ba' 'ab bc cc ca aa ac']
    >>> word = WordExtractor(window_size=2, numerosity_reduction=False)
    >>> print(word.transform(X))
    ['aa ab ba ab bb bb bb ba' 'ab bc cc cc cc ca aa ac']
    """

    def __init__(self, window_size=0.1, window_step=1,
                 numerosity_reduction=True):
        self.window_size = window_size
        self.window_step = window_step
        self.numerosity_reduction = numerosity_reduction

    def fit(self, X=None, y=None):
        """Pass.

        Parameters
        ----------
        X
            ignored

        y
            Ignored

        Returns
        -------
        self : object

        """
        return self

    def transform(self, X):
        """Transform time series into sequences of words.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)

        Returns
        -------
        X_new : array, shape = (n_samples,)
            Transformed data. Each row is a string consisting of words
            separated by a whitespace.

        """
        X = check_array(X, dtype=None)
        n_samples, n_timestamps = X.shape
        window_size, window_step = self._check_params(n_timestamps)
        n_windows = (n_timestamps - window_size + window_step) // window_step

        X_window = _windowed_view(X, n_samples, n_timestamps,
                                  window_size, window_step)
        X_word = np.asarray([[''.join(X_window[i, j])
                              for j in range(n_windows)]
                             for i in range(n_samples)])

        if self.numerosity_reduction:
            not_equal = np.c_[X_word[:, 1:] != X_word[:, :-1],
                              np.full(n_samples, True)]
            X_bow = np.asarray([' '.join(X_word[i, not_equal[i]])
                                for i in range(n_samples)])
        else:
            X_bow = np.asarray([' '.join(X_word[i]) for i in range(n_samples)])
        return X_bow

    def _check_params(self, n_timestamps):
        if not isinstance(self.window_size,
                          (int, np.integer, float, np.floating)):
            raise TypeError("'window_size' must be an integer or a float.")
        if isinstance(self.window_size, (int, np.integer)):
            if not 1 <= self.window_size <= n_timestamps:
                raise ValueError(
                    "If 'window_size' is an integer, it must be greater "
                    "than or equal to 1 and lower than or equal to "
                    "n_timestamps (got {0}).".format(self.window_size)
                )
            window_size = self.window_size
        else:
            if not 0 < self.window_size <= 1:
                raise ValueError(
                    "If 'window_size' is a float, it must be greater "
                    "than 0 and lower than or equal to 1 (got {0})."
                    .format(self.window_size)
                )
            window_size = ceil(self.window_size * n_timestamps)
        if not isinstance(self.window_step,
                          (int, np.integer, float, np.floating)):
            raise TypeError("'window_step' must be an integer or a float.")
        if isinstance(self.window_step, (int, np.integer)):
            if not 1 <= self.window_step <= n_timestamps:
                raise ValueError(
                    "If 'window_step' is an integer, it must be greater "
                    "than or equal to 1 and lower than or equal to "
                    "n_timestamps (got {0}).".format(self.window_step)
                )
            window_step = self.window_step
        else:
            if not 0 < self.window_step <= 1:
                raise ValueError(
                    "If 'window_step' is a float, it must be greater "
                    "than 0 and lower than or equal to 1 (got {0})."
                    .format(self.window_step)
                )
            window_step = ceil(self.window_step * n_timestamps)
        return window_size, window_step


class BagOfWords(BaseEstimator, TransformerMixin):
    """Bag-of-words representation for time series.

    This algorithm uses a sliding window to extract subsequences from the
    time series and transforms each subsequence into a word using the
    Piecewise Aggregate Approximation and the Symbolic Aggregate approXimation
    algorithms. Thus it transforms each time series into a bag of words.

    Parameters
    ----------
    window_size : int or float (default = 0.5)
        Length of the sliding window. If float, it represents
        a percentage of the size of each time series and must be
        between 0 and 1.

    word_size : int or float (default = 0.5)
        Length of the words. If float, it represents
        a percentage of the length of the sliding window and must be
        between 0. and 1.

    n_bins : int (default = 4)
        The number of bins to produce. It must be between 2 and
        ``min(n_timestamps, 26)``.

    strategy : 'uniform', 'quantile' or 'normal' (default = 'normal')
        Strategy used to define the widths of the bins:

        - 'uniform': All bins in each sample have identical widths
        - 'quantile': All bins in each sample have the same number of points
        - 'normal': Bin edges are quantiles from a standard normal distribution

    numerosity_reduction : bool (default = True)
        If True, delete sample-wise all but one occurence of back to back
        identical occurences of the same words.

    window_step : int or float (default = 1)
        Step of the sliding window. If float, it represents the percentage of
        the size of each time series and must be between 0 and 1. The window
        size will be computed as ``ceil(window_step * n_timestamps)``.

    threshold_std: int, float or None (default = 0.01)
        Threshold used to determine whether a subsequence is standardized.
        Subsequences whose standard deviations are lower than this threshold
        are not standardized. If None, all the subsequences are standardized.

    norm_mean : bool (default = True)
        If True, center each subseries before scaling.

    norm_std : bool (default = True)
        If True, scale each subseries to unit variance.

    overlapping : bool (default = True)
        If True, time points may belong to two bins when decreasing the size
        of the subsequence with the Piecewise Aggregate Approximation
        algorithm. If False, each time point belong to one single bin, but
        the size of the bins may vary.

    raise_warning : bool (default = False)
        If True, a warning is raised when the number of bins is smaller for
        at least one subsequence. In this case, you should consider decreasing
        the number of bins, using another strategy to compute the bins or
        removing the corresponding time series.

    alphabet : None or array-like, shape = (n_bins,)
        Alphabet to use. If None, the first `n_bins` letters of the Latin
        alphabet are used.

    References
    ----------
    .. [1] J. Lin, R. Khade and Y. Li, "Rotation-invariant similarity in time
           series using bag-of-patterns representation". Journal of Intelligent
           Information Systems, 39 (2), 287-315 (2012).

    Examples
    --------
    >>> import numpy as np
    >>> from pyts.bag_of_words import BagOfWords
    >>> X = np.arange(12).reshape(2, 6)
    >>> bow = BagOfWords(window_size=4, word_size=4)
    >>> bow.transform(X)
    array(['abcd', 'abcd'], dtype='<U4')
    >>> bow.set_params(numerosity_reduction=False)
    BagOfWords(...)
    >>> bow.transform(X)
    array(['abcd abcd abcd', 'abcd abcd abcd'], dtype='<U14')

    """

    def __init__(self, window_size=0.5, word_size=0.5, n_bins=4,
                 strategy='normal', numerosity_reduction=True, window_step=1,
                 threshold_std=0.01, norm_mean=True, norm_std=True,
                 overlapping=True, raise_warning=False, alphabet=None):
        self.window_size = window_size
        self.word_size = word_size
        self.n_bins = n_bins
        self.strategy = strategy
        self.numerosity_reduction = numerosity_reduction
        self.window_step = window_step
        self.threshold_std = threshold_std
        self.norm_mean = norm_mean
        self.norm_std = norm_std
        self.overlapping = overlapping
        self.raise_warning = raise_warning
        self.alphabet = alphabet

    def fit(self, X, y=None):
        """Pass.

        Parameters
        ----------
        X
            Ignored

        y
            Ignored

        Returns
        -------
        self : object

        """
        return self

    def transform(self, X):
        """Transform each time series into a bag of words.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Data to transform.

        Returns
        -------
        X_new : array, shape = (n_samples,)
            Bags of words.

        """
        X = check_array(X, dtype='float64')
        n_samples, n_timestamps = X.shape
        window_size, word_size, window_step, alphabet = self._check_params(
            n_timestamps)
        n_windows = (n_timestamps - window_size + window_step) // window_step

        # Standardize time series if quantile from standard normal distribution
        if self.strategy == 'normal':
            X_scaled = StandardScaler().transform(X)
        else:
            X_scaled = X

        # Extract subsequences using a sliding window
        X_window = _windowed_view(
            X_scaled, n_samples, n_timestamps, window_size, window_step
        ).reshape(n_samples * n_windows, window_size)

        if self.threshold_std is not None:

            # Identify subsequences whose standard deviation below threshold
            idx = np.std(X_window, axis=1) < self.threshold_std

            if np.any(idx):
                # Subsequences with standard deviations below threshold
                X_paa = PiecewiseAggregateApproximation(
                    window_size=None, output_size=word_size,
                    overlapping=self.overlapping
                ).transform(X_window[idx])

                # Compute the bin edges
                discretizer = KBinsDiscretizer(
                    n_bins=self.n_bins, strategy=self.strategy,
                    raise_warning=self.raise_warning
                )
                bin_edges = discretizer._compute_bins(
                    X_scaled, n_samples, self.n_bins, self.strategy)

                # Tile the bin edges for subsequences from the same time series
                if self.strategy != 'normal':
                    count = np.bincount(
                        np.floor_divide(np.nonzero(idx)[0], n_windows)
                    )
                    bin_edges = np.vstack([
                        np.tile(bin_edges[i], (count[i], 1))
                        for i in range(count.size) if count[i] != 0
                    ])

                X_sax_below_thresh = alphabet[_digitize(X_paa, bin_edges)]

        # Subsequences with standard deviations above threshold
        if (self.threshold_std is None) or (not np.all(idx)):
            pipeline = make_pipeline(
                StandardScaler(
                    with_mean=self.norm_mean, with_std=self.norm_std
                ),
                PiecewiseAggregateApproximation(
                    window_size=None, output_size=word_size,
                    overlapping=self.overlapping
                ),
                SymbolicAggregateApproximation(
                    n_bins=self.n_bins, strategy=self.strategy,
                    alphabet=self.alphabet, raise_warning=self.raise_warning
                )
            )
            if self.threshold_std is None:
                X_sax_above_thresh = pipeline.fit_transform(X_window)
            else:
                X_sax_above_thresh = pipeline.fit_transform(X_window[~idx])

        # Concatenate SAX words
        if self.threshold_std is not None:
            if np.any(idx):
                if not np.all(idx):
                    X_sax = np.empty((n_samples * n_windows, word_size),
                                     dtype='<U1')
                    X_sax[idx] = X_sax_below_thresh
                    X_sax[~idx] = X_sax_above_thresh
                else:
                    X_sax = X_sax_below_thresh
            else:
                X_sax = X_sax_above_thresh
        else:
            X_sax = X_sax_above_thresh
        X_sax = X_sax.reshape(n_samples, n_windows, word_size)

        # Join letters to make words
        X_word = np.asarray([[''.join(X_sax[i, j])
                              for j in range(n_windows)]
                             for i in range(n_samples)])

        # Apply numerosity reduction
        if self.numerosity_reduction:
            not_equal = np.c_[X_word[:, 1:] != X_word[:, :-1],
                              np.full(n_samples, True)]
            X_bow = np.asarray([' '.join(X_word[i, not_equal[i]])
                                for i in range(n_samples)])
        else:
            X_bow = np.asarray([' '.join(X_word[i]) for i in range(n_samples)])

        return X_bow

    def _check_params(self, n_timestamps):
        if not isinstance(self.window_size,
                          (int, np.integer, float, np.floating)):
            raise TypeError("'window_size' must be an integer or a float.")
        if isinstance(self.window_size, (int, np.integer)):
            if not 1 <= self.window_size <= n_timestamps:
                raise ValueError(
                    "If 'window_size' is an integer, it must be greater "
                    "than or equal to 1 and lower than or equal to "
                    "n_timestamps (got {0}).".format(self.window_size)
                )
            window_size = self.window_size
        else:
            if not 0 < self.window_size <= 1:
                raise ValueError(
                    "If 'window_size' is a float, it must be greater "
                    "than 0 and lower than or equal to 1 "
                    "(got {0}).".format(self.window_size)
                )
            window_size = ceil(self.window_size * n_timestamps)

        if not isinstance(self.word_size,
                          (int, np.integer, float, np.floating)):
            raise TypeError("'word_size' must be an integer or a float.")
        if isinstance(self.word_size, (int, np.integer)):
            if not 1 <= self.word_size <= window_size:
                raise ValueError(
                    "If 'word_size' is an integer, it must be greater "
                    "than or equal to 1 and lower than or equal to "
                    "window_size (got {0}).".format(self.word_size)
                )
            word_size = self.word_size
        else:
            if not 0 < self.word_size <= 1:
                raise ValueError(
                    "If 'word_size' is a float, it must be greater "
                    "than 0 and lower than or equal to 1 "
                    "(got {0}).".format(self.word_size)
                )
            word_size = ceil(self.word_size * window_size)

        if not isinstance(self.n_bins, (int, np.integer)):
            raise TypeError("'n_bins' must be an integer.")
        if not 2 <= self.n_bins <= 26:
            raise ValueError(
                "'n_bins' must be greater than or equal to 2 and lower than "
                "or equal to 26 (got {0})."
                .format(self.n_bins)
            )

        if self.strategy not in ['uniform', 'quantile', 'normal']:
            raise ValueError("'strategy' must be either 'uniform', 'quantile' "
                             "or 'normal' (got {0})".format(self.strategy))

        if not isinstance(self.window_step,
                          (int, np.integer, float, np.floating)):
            raise TypeError("'window_step' must be an integer or a float.")
        if isinstance(self.window_step, (int, np.integer)):
            if not 1 <= self.window_step <= n_timestamps:
                raise ValueError(
                    "If 'window_step' is an integer, it must be greater "
                    "than or equal to 1 and lower than or equal to "
                    "n_timestamps (got {0}).".format(self.window_step)
                )
            window_step = self.window_step
        else:
            if not 0 < self.window_step <= 1:
                raise ValueError(
                    "If 'window_step' is a float, it must be greater "
                    "than 0 and lower than or equal to 1 (got {0})."
                    .format(self.window_step)
                )
            window_step = ceil(self.window_step * n_timestamps)

        threshold_std_none = self.threshold_std is None
        threshold_std_int_float = isinstance(
            self.threshold_std, (int, np.integer, float, np.floating))
        if not (threshold_std_none or threshold_std_int_float):
            raise TypeError(
                "'threshold_std' must be an integer, a float or None."
            )
        if threshold_std_int_float and (not self.threshold_std >= 0.):
            raise ValueError("If 'threshold_std' is an integer or a float, it "
                             "must be non-negative (got {0})."
                             .format(self.threshold_std))

        if not ((self.alphabet is None)
                or (isinstance(self.alphabet, (list, tuple, np.ndarray)))):
            raise TypeError("'alphabet' must be None or array-like "
                            "with shape (n_bins,) (got {0})."
                            .format(self.alphabet))
        if self.alphabet is None:
            alphabet = np.array([chr(i) for i in range(97, 97 + self.n_bins)])
        else:
            alphabet = check_array(self.alphabet, ensure_2d=False, dtype=None)
            if alphabet.shape != (self.n_bins,):
                raise ValueError("If 'alphabet' is array-like, its shape "
                                 "must be equal to (n_bins,).")

        return window_size, word_size, window_step, alphabet



"""Testing for BOSS metric."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
import re
from math import sqrt
from pyts.metrics import boss


x = np.arange(1, 6)
y = np.arange(1, 6)[::-1]
z = [0, 0, 0, 10, 0]


@pytest.mark.parametrize(
    'x, y, err_msg',
    [(x.reshape(1, -1), y, "'x' must a one-dimensional array."),
     (x, y.reshape(1, -1), "'y' must a one-dimensional array."),
     (x[:2], y, "'x' and 'y' must have the same shape.")]
)
def test_parameter_check(x, y, err_msg):
    """Test parameter validation."""
    with pytest.raises(ValueError, match=re.escape(err_msg)):
        boss(x, y)


@pytest.mark.parametrize(
    'x, y, arr_desired',
    [(x, y, sqrt(np.sum((x - y) ** 2))),
     (y, x, sqrt(np.sum((x - y) ** 2))),
     (x, z, sqrt(np.sum((x - z) ** 2))),
     (z, x, 6),
     (y, z, sqrt(np.sum((y - z) ** 2))),
     (z, y, 8)]
)
def test_actual_results(x, y, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = boss(x, y)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)



"""The :mod:`pyts.metrics` module includes metrics."""

from .boss import boss
from .dtw import dtw, itakura_parallelogram, sakoe_chiba_band, show_options
from .lower_bounds import (lower_bound_improved, lower_bound_keogh,
                           lower_bound_kim, lower_bound_yi)


__all__ = ['boss', 'dtw', 'itakura_parallelogram', 'lower_bound_improved',
           'lower_bound_keogh', 'lower_bound_kim', 'lower_bound_yi',
           'sakoe_chiba_band', 'show_options']



"""The :mod:`pyts.preprocessing` module includes preprocessing algorithms."""

from .scaler import StandardScaler, MinMaxScaler, MaxAbsScaler, RobustScaler
from .transformer import PowerTransformer, QuantileTransformer
from .discretizer import KBinsDiscretizer
from .imputer import InterpolationImputer


__all__ = ['StandardScaler',
           'MinMaxScaler',
           'MaxAbsScaler',
           'RobustScaler',
           'PowerTransformer',
           'QuantileTransformer',
           'KBinsDiscretizer',
           'InterpolationImputer']



"""Code for Bag-of-SFA Symbols in Vector Space."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
from math import ceil
from sklearn.utils.validation import check_array, check_X_y, check_is_fitted
from sklearn.utils.multiclass import check_classification_targets
from sklearn.base import BaseEstimator
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.preprocessing import LabelEncoder
from sklearn.feature_extraction.text import TfidfVectorizer
from ..approximation import SymbolicFourierApproximation
from ..base import UnivariateClassifierMixin
from ..utils.utils import _windowed_view


class BOSSVS(BaseEstimator, UnivariateClassifierMixin):
    """Bag-of-SFA Symbols in Vector Space.

    Each time series is transformed into an histogram using the
    Bag-of-SFA Symbols (BOSS) algorithm. Then, for each class, the histograms
    are added up and a tf-idf vector is computed. The predicted class for
    a new sample is the class giving the highest cosine similarity between
    its tf vector and the tf-idf vectors of each class.

    Parameters
    ----------
    word_size : int (default = 4)
        Size of each word.

    n_bins : int (default = 4)
        The number of bins to produce. It must be between 2 and 26.

    window_size : int or float (default = 10)
        Size of the sliding window. If float, it represents the percentage of
        the size of each time series and must be between 0 and 1. The window
        size will be computed as ``ceil(window_size * n_timestamps)``.

    window_step : int or float (default = 1)
        Step of the sliding window. If float, it represents the percentage of
        the size of each time series and must be between 0 and 1. The window
        size will be computed as ``ceil(window_step * n_timestamps)``.

    anova : bool (default = False)
        If True, the Fourier coefficient selection is done via a one-way
        ANOVA test. If False, the first Fourier coefficients are selected.

    drop_sum : bool (default = False)
        If True, the first Fourier coefficient (i.e. the sum of the subseries)
        is dropped. Otherwise, it is kept.

    norm_mean : bool (default = False)
        If True, center each subseries before scaling.

    norm_std : bool (default = False)
        If True, scale each subseries to unit variance.

    strategy : str (default = 'quantile')
        Strategy used to define the widths of the bins:

        - 'uniform': All bins in each sample have identical widths
        - 'quantile': All bins in each sample have the same number of points
        - 'normal': Bin edges are quantiles from a standard normal distribution
        - 'entropy': Bin edges are computed using information gain

    alphabet : None, 'ordinal' or array-like, shape = (n_bins,)
        Alphabet to use. If None, the first `n_bins` letters of the Latin
        alphabet are used.

    numerosity_reduction : bool (default = True)
        If True, delete sample-wise all but one occurence of back to back
        identical occurences of the same words.

    use_idf : bool (default = True)
        Enable inverse-document-frequency reweighting.

    smooth_idf : bool (default = False)
        Smooth idf weights by adding one to document frequencies, as if an
        extra document was seen containing every term in the collection
        exactly once. Prevents zero divisions.

    sublinear_tf : bool (default = True)
        Apply sublinear tf scaling, i.e. replace tf with 1 + log(tf).

    Attributes
    ----------
    classes_ : array, shape = (n_classes,)
        An array of class labels known to the classifier.

    idf_ : array, shape = (n_features,) , or None
        The learned idf vector (global term weights) when ``use_idf=True``,
        None otherwise.

    tfidf_ : array, shape = (n_classes, n_words)
        Term-document matrix.

    vocabulary_ : dict
        A mapping of feature indices to terms.

    References
    ----------
    .. [1] P. Schäfer, "Scalable Time Series Classification". Data Mining and
           Knowledge Discovery, 30(5), 1273-1298 (2016).

    Examples
    --------
    >>> from pyts.classification import BOSSVS
    >>> from pyts.datasets import load_gunpoint
    >>> X_train, X_test, y_train, y_test = load_gunpoint(return_X_y=True)
    >>> clf = BOSSVS(window_size=28)
    >>> clf.fit(X_train, y_train)
    BOSSVS(...)
    >>> clf.score(X_test, y_test)
    0.98

    """

    def __init__(self, word_size=4, n_bins=4, window_size=10, window_step=1,
                 anova=False, drop_sum=False, norm_mean=False, norm_std=False,
                 strategy='quantile', alphabet=None, numerosity_reduction=True,
                 use_idf=True, smooth_idf=False, sublinear_tf=True):
        self.word_size = word_size
        self.n_bins = n_bins
        self.window_size = window_size
        self.window_step = window_step
        self.anova = anova
        self.drop_sum = drop_sum
        self.norm_mean = norm_mean
        self.norm_std = norm_std
        self.strategy = strategy
        self.alphabet = alphabet
        self.numerosity_reduction = numerosity_reduction
        self.use_idf = use_idf
        self.smooth_idf = smooth_idf
        self.sublinear_tf = sublinear_tf

    def fit(self, X, y):
        """Compute the document-term matrix.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Training vector.

        y : array-like, shape = (n_samples,)
            Class labels for each data sample.

        Returns
        -------
        self : object

        """
        X, y = check_X_y(X, y)
        n_samples, n_timestamps = X.shape
        check_classification_targets(y)
        le = LabelEncoder()
        y_ind = le.fit_transform(y)
        self.classes_ = le.classes_
        n_classes = self.classes_.size

        window_size, window_step = self._check_params(n_timestamps)
        n_windows = (n_timestamps - window_size + window_step) // window_step

        X_windowed = _windowed_view(
            X, n_samples, n_timestamps, window_size, window_step
        )
        X_windowed = X_windowed.reshape(n_samples * n_windows, window_size)

        sfa = SymbolicFourierApproximation(
            n_coefs=self.word_size, drop_sum=self.drop_sum, anova=self.anova,
            norm_mean=self.norm_mean, norm_std=self.norm_std,
            n_bins=self.n_bins, strategy=self.strategy, alphabet=self.alphabet
        )
        y_repeated = np.repeat(y, n_windows)
        X_sfa = sfa.fit_transform(X_windowed, y_repeated)

        X_word = np.asarray([''.join(X_sfa[i])
                             for i in range(n_samples * n_windows)])
        X_word = X_word.reshape(n_samples, n_windows)

        if self.numerosity_reduction:
            not_equal = np.c_[X_word[:, 1:] != X_word[:, :-1],
                              np.full(n_samples, True)]
            X_bow = np.asarray([' '.join(X_word[i, not_equal[i]])
                                for i in range(n_samples)])
        else:
            X_bow = np.asarray([' '.join(X_word[i]) for i in range(n_samples)])

        X_class = np.array([' '.join(X_bow[y_ind == i])
                            for i in range(n_classes)])

        tfidf = TfidfVectorizer(
            norm=None, use_idf=self.use_idf, smooth_idf=self.smooth_idf,
            sublinear_tf=self.sublinear_tf
        )
        self.tfidf_ = tfidf.fit_transform(X_class).toarray()
        self.vocabulary_ = {value: key for key, value in
                            tfidf.vocabulary_.items()}
        if self.use_idf:
            self.idf_ = tfidf.idf_
        else:
            self.idf_ = None

        self._window_size = window_size
        self._window_step = window_step
        self._n_windows = n_windows
        self._tfidf = tfidf
        self._sfa = sfa
        return self

    def decision_function(self, X):
        """Evaluate the cosine similarity between document-term matrix and X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_timestamps)
            Test samples.

        Returns
        -------
        X : array, shape (n_samples, n_classes)
            Cosine similarity between the document-term matrix and X.

        """
        check_is_fitted(self, ['vocabulary_', 'tfidf_', 'idf_', '_tfidf'])
        X = check_array(X, dtype='float64')
        n_samples, n_timestamps = X.shape

        X_windowed = _windowed_view(
            X, n_samples, n_timestamps, self._window_size, self._window_step
        )
        X_windowed = X_windowed.reshape(-1, self._window_size)

        X_sfa = self._sfa.transform(X_windowed)
        X_word = np.asarray([''.join(X_sfa[i]) for i in range(X_sfa.shape[0])])
        X_word = X_word.reshape(n_samples, self._n_windows)

        if self.numerosity_reduction:
            not_equal = np.c_[X_word[:, 1:] != X_word[:, :-1],
                              np.full(n_samples, True)]
            X_bow = np.asarray([' '.join(X_word[i, not_equal[i]])
                                for i in range(n_samples)])
        else:
            X_bow = np.asarray([' '.join(X_word[i]) for i in range(n_samples)])

        X_tf = self._tfidf.transform(X_bow).toarray()
        if self.idf_ is not None:
            X_tf /= self.idf_
        return cosine_similarity(X_tf, self.tfidf_)

    def predict(self, X):
        """Predict the class labels for the provided data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Test samples.

        Returns
        -------
        y_pred : array, shape = (n_samples,)
            Class labels for each data sample.

        """
        return self.classes_[self.decision_function(X).argmax(axis=1)]

    def _check_params(self, n_timestamps):
        if not isinstance(self.word_size, (int, np.integer)):
            raise TypeError("'word_size' must be an integer.")
        if not self.word_size >= 1:
            raise ValueError("'word_size' must be a positive integer.")

        if not isinstance(self.window_size,
                          (int, np.integer, float, np.floating)):
            raise TypeError("'window_size' must be an integer or a float.")
        if isinstance(self.window_size, (int, np.integer)):
            if self.drop_sum:
                if not 1 <= self.window_size <= (n_timestamps - 1):
                    raise ValueError(
                        "If 'window_size' is an integer, it must be greater "
                        "than or equal to 1 and lower than or equal to "
                        "(n_timestamps - 1) if 'drop_sum=True'."
                    )
            else:
                if not 1 <= self.window_size <= n_timestamps:
                    raise ValueError(
                        "If 'window_size' is an integer, it must be greater "
                        "than or equal to 1 and lower than or equal to "
                        "n_timestamps if 'drop_sum=False'."
                    )
            window_size = self.window_size
        else:
            if not 0 < self.window_size <= 1:
                raise ValueError(
                    "If 'window_size' is a float, it must be greater "
                    "than 0 and lower than or equal to 1."
                )
            window_size = ceil(self.window_size * n_timestamps)

        if not isinstance(self.window_step,
                          (int, np.integer, float, np.floating)):
            raise TypeError("'window_step' must be an integer or a float.")
        if isinstance(self.window_step, (int, np.integer)):
            if not 1 <= self.window_step <= n_timestamps:
                raise ValueError(
                    "If 'window_step' is an integer, it must be greater "
                    "than or equal to 1 and lower than or equal to "
                    "n_timestamps."
                )
            window_step = self.window_step
        else:
            if not 0 < self.window_step <= 1:
                raise ValueError(
                    "If 'window_step' is a float, it must be greater "
                    "than 0 and lower than or equal to 1."
                )
            window_step = ceil(self.window_step * n_timestamps)

        if self.drop_sum:
            if not self.word_size <= (window_size - 1):
                raise ValueError(
                    "'word_size' must be lower than or equal to "
                    "(window_size - 1) if 'drop_sum=True'."
                )
        else:
            if not self.word_size <= window_size:
                raise ValueError(
                    "'word_size' must be lower than or equal to "
                    "window_size if 'drop_sum=False'."
                )
        return window_size, window_step



"""Code for SAX-VSM."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from sklearn.utils.multiclass import check_classification_targets
from sklearn.base import BaseEstimator
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.preprocessing import LabelEncoder
from sklearn.feature_extraction.text import CountVectorizer, TfidfVectorizer
from ..bag_of_words import BagOfWords
from ..base import UnivariateClassifierMixin


class SAXVSM(BaseEstimator, UnivariateClassifierMixin):
    """Classifier based on SAX-VSM representation and tf-idf statistics.

    Time series are first transformed into bag of words using Symbolic
    Aggregate approXimation (SAX) algorithm followed by a bag-of-words
    model. Then the classes are transformed into a Vector Space Model
    (VSM) using term frequencies (tf) and inverse document frequencies
    (idf).

    Parameters
    ----------
    window_size : int or float (default = 0.5)
        Length of the sliding window. If float, it represents
        a percentage of the size of each time series and must be
        between 0 and 1.

    word_size : int or float (default = 0.5)
        Length of the words. If float, it represents
        a percentage of the length of the sliding window and must be
        between 0. and 1.

    n_bins : int (default = 4)
        The number of bins to produce. It must be between 2 and
        ``min(window_size, 26)``.

    strategy : 'uniform', 'quantile' or 'normal' (default = 'normal')
        Strategy used to define the widths of the bins:

        - 'uniform': All bins in each sample have identical widths
        - 'quantile': All bins in each sample have the same number of points
        - 'normal': Bin edges are quantiles from a standard normal distribution

    numerosity_reduction : bool (default = True)
        If True, delete sample-wise all but one occurence of back to back
        identical occurences of the same words.

    window_step : int or float (default = 1)
        Step of the sliding window. If float, it represents the percentage of
        the size of each time series and must be between 0 and 1. The step of
        sliding window will be computed as
        ``ceil(window_step * n_timestamps)``.

    threshold_std: float (default = 0.01)
        Threshold used to determine whether a subsequence is standardized.
        Subsequences whose standard deviations are lower than this threshold
        are not standardized.

    norm_mean : bool (default = True)
        If True, center each subseries before scaling.

    norm_std : bool (default = True)
        If True, scale each subseries to unit variance.

    use_idf : bool (default = True)
        Enable inverse-document-frequency reweighting.

    smooth_idf : bool (default = False)
        Smooth idf weights by adding one to document frequencies, as if an
        extra document was seen containing every term in the collection
        exactly once. Prevents zero divisions.

    sublinear_tf : bool (default = True)
        Apply sublinear tf scaling, i.e. replace tf with 1 + log(tf).

    overlapping : bool (default = True)
        If True, time points may belong to two bins when decreasing the size
        of the subsequence with the Piecewise Aggregate Approximation
        algorithm. If False, each time point belong to one single bin, but
        the size of the bins may vary.

    alphabet : None or array-like, shape = (n_bins,)
        Alphabet to use. If None, the first `n_bins` letters of the Latin
        alphabet are used.

    Attributes
    ----------
    classes_ : array, shape = (n_classes,)
        An array of class labels known to the classifier.

    idf_ : array, shape = (n_features,) , or None
        The learned idf vector (global term weights) when ``use_idf=True``,
        None otherwise.

    tfidf_ : array, shape = (n_classes, n_words)
        Term-document matrix.

    vocabulary_ : dict
        A mapping of feature indices to terms.

    References
    ----------
    .. [1] P. Senin, and S. Malinchik, "SAX-VSM: Interpretable Time Series
           Classification Using SAX and Vector Space Model". International
           Conference on Data Mining, 13, 1175-1180 (2013).

    Examples
    --------
    >>> from pyts.classification import SAXVSM
    >>> from pyts.datasets import load_gunpoint
    >>> X_train, X_test, y_train, y_test = load_gunpoint(return_X_y=True)
    >>> clf = SAXVSM(window_size=64, word_size=12, n_bins=5, strategy='normal')
    >>> clf.fit(X_train, y_train)
    SAXVSM(...)
    >>> clf.score(X_test, y_test)
    1.0

    """

    def __init__(self, window_size=0.5, word_size=0.5, n_bins=4,
                 strategy='normal', numerosity_reduction=True, window_step=1,
                 threshold_std=0.01, norm_mean=True, norm_std=True,
                 use_idf=True, smooth_idf=False, sublinear_tf=True,
                 overlapping=True, alphabet=None):
        self.window_size = window_size
        self.word_size = word_size
        self.n_bins = n_bins
        self.strategy = strategy
        self.numerosity_reduction = numerosity_reduction
        self.window_step = window_step
        self.threshold_std = threshold_std
        self.norm_mean = norm_mean
        self.norm_std = norm_std
        self.use_idf = use_idf
        self.smooth_idf = smooth_idf
        self.sublinear_tf = sublinear_tf
        self.overlapping = overlapping
        self.alphabet = alphabet

    def fit(self, X, y):
        """Fit the model according to the given training data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Training vector.

        y : array-like, shape = (n_samples,)
            Class labels for each data sample.

        Returns
        -------
        self : object

        """
        X, y = check_X_y(X, y)
        check_classification_targets(y)
        le = LabelEncoder()
        y_ind = le.fit_transform(y)
        self.classes_ = le.classes_
        n_classes = self.classes_.size

        # Transform each time series into a bag of words
        bow = BagOfWords(
            window_size=self.window_size, word_size=self.word_size,
            n_bins=self.n_bins, strategy=self.strategy,
            numerosity_reduction=self.numerosity_reduction,
            window_step=self.window_step, threshold_std=self.threshold_std,
            norm_mean=self.norm_mean, norm_std=self.norm_std,
            overlapping=self.overlapping, alphabet=self.alphabet
        )
        X_bow = bow.fit_transform(X)

        X_class = [' '.join(X_bow[y_ind == classe])
                   for classe in range(n_classes)]

        tfidf = TfidfVectorizer(
            norm=None, use_idf=self.use_idf, smooth_idf=self.smooth_idf,
            sublinear_tf=self.sublinear_tf
        )
        self.tfidf_ = tfidf.fit_transform(X_class).A
        self.vocabulary_ = {value: key for key, value in
                            tfidf.vocabulary_.items()}
        if self.use_idf:
            self.idf_ = tfidf.idf_
        else:
            self.idf_ = None
        self._tfidf = tfidf
        self._bow = bow
        return self

    def decision_function(self, X):
        """Evaluate the cosine similarity between document-term matrix and X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_timestamps)
            Test samples.

        Returns
        -------
        X : array-like, shape (n_samples, n_classes)
            osine similarity between the document-term matrix and X.

        """
        check_is_fitted(self, ['vocabulary_', 'tfidf_', 'idf_',
                               '_tfidf', 'classes_'])
        X = check_array(X)
        X_bow = self._bow.transform(X)
        vectorizer = CountVectorizer(vocabulary=self._tfidf.vocabulary_)
        X_transformed = vectorizer.transform(X_bow).toarray()
        return cosine_similarity(X_transformed, self.tfidf_)

    def predict(self, X):
        """Predict the class labels for the provided data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Test samples.

        Returns
        -------
        y_pred : array-like, shape = (n_samples,)
            Class labels for each data sample.

        """
        return self.classes_[self.decision_function(X).argmax(axis=1)]



"""Testing for Symbolic Aggregate Approximation."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
import re
from pyts.approximation import SymbolicAggregateApproximation


X = np.arange(30).reshape(3, 10)


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'n_bins': '3'}, TypeError, "'n_bins' must be an integer."),

     ({'alphabet': 'whoops'}, TypeError,
      "'alphabet' must be None, 'ordinal' or array-like with shape (n_bins,) "
      "(got {0})".format('whoops')),

     ({'n_bins': 1}, ValueError,
      "'n_bins' must be greater than or equal to 2 and lower than "
      "or equal to 26 (got 1)."),

     ({'n_bins': 27}, ValueError,
      "'n_bins' must be greater than or equal to 2 and lower than "
      "or equal to 26 (got 27)."),

     ({'strategy': 'whoops'}, ValueError,
      "'strategy' must be either 'uniform', 'quantile' or 'normal' "
      "(got {0})".format('whoops')),

     ({'alphabet': ['a', 'b', 'c']}, ValueError,
      "If 'alphabet' is array-like, its shape must be equal to (n_bins,).")]
)
def test_parameter_check(params, error, err_msg):
    """Test parameter validation."""
    sax = SymbolicAggregateApproximation(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        sax.transform(X)


@pytest.mark.parametrize(
    'params, X, arr_desired',
    [({}, [[0, 1, 2, 3]], [['a', 'b', 'c', 'd']]),

     ({'strategy': 'uniform'}, [[0, 1, 2, 3]], [['a', 'b', 'c', 'd']]),

     ({}, [[0, 4, 2, 6]], [['a', 'c', 'b', 'd']]),

     ({}, [[-5, -8, -7, -6]], [['d', 'a', 'b', 'c']]),

     ({'alphabet': 'ordinal'}, [[0, 1, 2, 3]], [[0, 1, 2, 3]]),

     ({'alphabet': ['d', 'c', 'b', 'a']}, [[0, 1, 2, 3]],
      [['d', 'c', 'b', 'a']]),

     ({'alphabet': ['0', '1', '2', '3']}, [[0, 3, 2, 1]],
      [['0', '3', '2', '1']])]
)
def test_actual_results(params, X, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = SymbolicAggregateApproximation(**params).fit_transform(X)
    np.testing.assert_array_equal(arr_actual, arr_desired)



"""Joint Recurrence Plots."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
from sklearn.base import BaseEstimator
from ...base import MultivariateTransformerMixin
from ...image import RecurrencePlot
from ..utils import check_3d_array


class JointRecurrencePlot(BaseEstimator, MultivariateTransformerMixin):
    r"""Joint Recurrence Plot.

    A recurrence plot is an image representing the distances between
    trajectories extracted from the original time series.

    A joint recurrence plot is an extension of recurrence plots for
    multivariate time series: it is the Hadamard of the recurrence
    plots obtained for each feature of the multivariate time series.

    Parameters
    ----------
    dimension : int or float (default = 1)
        Dimension of the trajectory. If float, it represents a percentage of
        the size of each time series and must be between 0 and 1.

    time_delay : int or float (default = 1)
        Time gap between two back-to-back points of the trajectory. If
        float, it represents a percentage of the size of each time series and
        must be between 0 and 1.

    threshold : float, 'point', 'distance' or None or list thereof (default = None)
        Threshold for the minimum distance. If None, the recurrence plots
        are not binarized. If ``threshold='point'``, the threshold is computed
        such as ``percentage`` percents of the points are smaller than the
        threshold. If ``threshold='distance'``, the threshold is computed as
        the ``percentage`` of the maximum distance.

    percentage : int, float or list thereof (default = 10)
        Percentage of black points if ``threshold='point'`` or percentage of
        maximum distance for threshold if ``threshold='distance'``.
        Ignored if ``threshold`` is a float or None. Note that the percentage
        is calculated for each recurrence plot independently, which implies
        that there will probably be less than `percentage` percents of black
        points in the joint recurrence plot.

    References
    ----------
    .. [1] M. Romano, M. Thiel, J. Kurths and W. con Bloh, "Multivariate
           Recurrence Plots". Physics Letters A (2004)

    Examples
    --------
    >>> from pyts.datasets import load_basic_motions
    >>> from pyts.multivariate.image import JointRecurrencePlot
    >>> X, _, _, _ = load_basic_motions(return_X_y=True)
    >>> transformer = JointRecurrencePlot()
    >>> X_new = transformer.transform(X)
    >>> X_new.shape
    (40, 100, 100)

    """  # noqa: E501

    def __init__(self, dimension=1, time_delay=1, threshold=None,
                 percentage=10):
        self.dimension = dimension
        self.time_delay = time_delay
        self.threshold = threshold
        self.percentage = percentage

    def fit(self, X=None, y=None):
        """Pass.

        Parameters
        ----------
        X
            Ignored
        y
            Ignored

        Returns
        -------
        self : object

        """
        return self

    def transform(self, X):
        """Transform each time series into a joint recurrence plot.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features, n_timestamps)
            Multivariate time series.

        Returns
        -------
        X_new : array, shape = (n_samples, image_size, image_size)
            Joint Recurrence plots. ``image_size`` is the number of
            trajectories and is equal to
            ``n_timestamps - (dimension - 1) * time_delay``.

        """
        X = check_3d_array(X)
        _, n_features, _ = X.shape
        thresholds_, percentages_ = self._check_params(n_features)

        X_rp = [self._joint_recurrence_plot(
            X[:, i, :], self.dimension, self.time_delay,
            thresholds_[i], percentages_[i]) for i in range(n_features)]
        X_jrp = np.product(X_rp, axis=0)
        return X_jrp

    @staticmethod
    def _joint_recurrence_plot(X, dimension, time_delay,
                               threshold, percentage):
        recurrence_plot = RecurrencePlot(
            dimension, time_delay, threshold, percentage)
        return recurrence_plot.transform(X)

    def _check_params(self, n_features):
        if isinstance(self.threshold, (tuple, list, np.ndarray)):
            if len(self.threshold) != n_features:
                raise ValueError(
                    "If 'threshold' is a list, its length must be equal to "
                    "n_features ({0} != {1})."
                    .format(len(self.threshold), n_features)
                )
            thresholds_ = self.threshold
        else:
            thresholds_ = [self.threshold for _ in range(n_features)]
        if isinstance(self.percentage, (tuple, list, np.ndarray)):
            if len(self.percentage) != n_features:
                raise ValueError(
                    "If 'percentage' is a list, its length must be equal to "
                    "n_features ({0} != {1})."
                    .format(len(self.percentage), n_features)
                )
            percentages_ = self.percentage
        else:
            percentages_ = [self.percentage for _ in range(n_features)]
        return thresholds_, percentages_



"""Base classes for all estimators."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

from sklearn.metrics import accuracy_score


class UnivariateTransformerMixin:
    """Mixin class for all univariate transformers in pyts."""

    def fit_transform(self, X, y=None, **fit_params):
        """Fit to data, then transform it.

        Fits transformer to `X` and `y` with optional parameters `fit_params`
        and returns a transformed version of `X`.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Univariate time series.

        y : None or array-like, shape = (n_samples,) (default = None)
            Target values (None for unsupervised transformations).

        **fit_params : dict
            Additional fit parameters.

        Returns
        -------
        X_new : array
            Transformed array.

        """  # noqa: E501
        if y is None:
            # fit method of arity 1 (unsupervised transformation)
            return self.fit(X, **fit_params).transform(X)
        else:
            # fit method of arity 2 (supervised transformation)
            return self.fit(X, y, **fit_params).transform(X)


class MultivariateTransformerMixin:
    """Mixin class for all multivariate transformers in pyts."""

    def fit_transform(self, X, y=None, **fit_params):
        """Fit to data, then transform it.

        Fits transformer to `X` and `y` with optional parameters `fit_params`
        and returns a transformed version of `X`.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features, n_timestamps)
            Multivariate time series.

        y : None or array-like, shape = (n_samples,) (default = None)
            Target values (None for unsupervised transformations).

        **fit_params : dict
            Additional fit parameters.

        Returns
        -------
        X_new : array
            Transformed array.

        """  # noqa: E501
        if y is None:
            # fit method of arity 1 (unsupervised transformation)
            return self.fit(X, **fit_params).transform(X)
        else:
            # fit method of arity 2 (supervised transformation)
            return self.fit(X, y, **fit_params).transform(X)


class UnivariateClassifierMixin:
    """Mixin class for all univariate classifiers in pyts."""

    _estimator_type = "classifier"

    def score(self, X, y, sample_weight=None):
        """
        Return the mean accuracy on the given test data and labels.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Univariate time series.

        y : array-like, shape = (n_samples,)
            True labels for `X`.

        sample_weight : None or array-like, shape = (n_samples,) (default = None)
            Sample weights.

        Returns
        -------
        score : float
            Mean accuracy of ``self.predict(X)`` with regards to `y`.

        """  # noqa: E501
        return accuracy_score(y, self.predict(X), sample_weight=sample_weight)


class MultivariateClassifierMixin:
    """Mixin class for all multivariate classifiers in pyts."""

    _estimator_type = "classifier"

    def score(self, X, y, sample_weight=None):
        """
        Return the mean accuracy on the given test data and labels.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features, n_timestamps)
            Multivariate time series.

        y : array-like, shape = (n_samples,)
            True labels for `X`.

        sample_weight : None or array-like, shape = (n_samples,) (default = None)
            Sample weights.

        Returns
        -------
        score : float
            Mean accuracy of ``self.predict(X)`` with regards to `y`.

        """  # noqa: E501
        return accuracy_score(y, self.predict(X), sample_weight=sample_weight)



"""Code for scalers."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import StandardScaler as SklearnStandardScaler
from sklearn.preprocessing import MinMaxScaler as SklearnMinMaxScaler
from sklearn.preprocessing import MaxAbsScaler as SklearnMaxAbsScaler
from sklearn.preprocessing import RobustScaler as SklearnRobustScaler
from sklearn.utils.validation import check_array
from ..base import UnivariateTransformerMixin


class StandardScaler(BaseEstimator, UnivariateTransformerMixin):
    """Standardize time series by removing mean and scaling to unit variance.

    Parameters
    ----------
    with_mean : bool (default = True)
        If True, center the data before scaling.

    with_std : bool (default = True)
        If True, scale the data to unit variance.

    Examples
    --------
    >>> from pyts.preprocessing import StandardScaler
    >>> X = [[0, 2, 0, 4, 4, 6, 4, 4],
    ...      [1, 0, 3, 2, 2, 2, 0, 2]]
    >>> scaler = StandardScaler()
    >>> scaler.transform(X)
    array([[-1.5, -0.5, -1.5,  0.5,  0.5,  1.5,  0.5,  0.5],
           [-0.5, -1.5,  1.5,  0.5,  0.5,  0.5, -1.5,  0.5]])

    """

    def __init__(self, with_mean=True, with_std=True):
        self.with_mean = with_mean
        self.with_std = with_std

    def fit(self, X=None, y=None):
        """Pass.

        Parameters
        ----------
        X
            Ignored

        y
            Ignored

        Returns
        -------
        self : object

        """
        return self

    def transform(self, X):
        """Perform standardization by centering and scaling.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Data to scale.

        Returns
        -------
        X_new : array-like, shape = (n_samples, n_timestamps)
            Scaled data.

        """
        X = check_array(X, dtype='float64')
        scaler = SklearnStandardScaler(
            with_mean=self.with_mean, with_std=self.with_std)
        X_new = scaler.fit_transform(X.T).T
        return X_new


class MinMaxScaler(BaseEstimator, TransformerMixin):
    """Transforms samples by scaling each sample to a given range.

    Parameters
    ----------
    sample_range : tuple (min, max) (default = (0, 1))
        Desired range of transformed data.

    Examples
    --------
    >>> from pyts.preprocessing import MinMaxScaler
    >>> X = [[1, 5, 3, 2, 9, 6, 4, 7],
    ...      [1, -2, 3, 2, 2, 1, 0, 2]]
    >>> scaler = MinMaxScaler()
    >>> scaler.transform(X)
    array([[0.   , 0.5  , 0.25 , 0.125, 1.   , 0.625, 0.375, 0.75 ],
           [0.6  , 0.   , 1.   , 0.8  , 0.8  , 0.6  , 0.4  , 0.8  ]])

    """

    def __init__(self, sample_range=(0, 1)):
        self.sample_range = sample_range

    def fit(self, X=None, y=None):
        """Pass.

        Parameters
        ----------
        X
            Ignored

        y
            Ignored

        Returns
        -------
        self : object

        """
        return self

    def transform(self, X):
        """Scale samples of X according to sample_range.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Data to scale.

        Returns
        -------
        X_new : array-like, shape = (n_samples, n_timestamps)
            Scaled data.

        """
        X = check_array(X, dtype='float64')
        scaler = SklearnMinMaxScaler(feature_range=self.sample_range)
        X_new = scaler.fit_transform(X.T).T
        return X_new


class MaxAbsScaler(BaseEstimator, TransformerMixin):
    """Scale each sample by its maximum absolute value.

    Examples
    --------
    >>> from pyts.preprocessing import MaxAbsScaler
    >>> X = [[1, 5, 3, 2, 10, 6, 4, 7],
    ...      [1, -5, 3, 2, 2, 1, 0, 2]]
    >>> scaler = MaxAbsScaler()
    >>> scaler.transform(X)
    array([[ 0.1,  0.5,  0.3,  0.2,  1. ,  0.6,  0.4,  0.7],
           [ 0.2, -1. ,  0.6,  0.4,  0.4,  0.2,  0. ,  0.4]])

    """

    def __init__(self):
        pass

    def fit(self, X=None, y=None):
        """Pass.

        Parameters
        ----------
        X
            Ignored

        y
            Ignored

        Returns
        -------
        self : object

        """
        return self

    def transform(self, X):
        """Scale the data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Data to scale.

        Returns
        -------
        X_new : array-like, shape = (n_samples, n_timestamps)
            Scaled data.

        """
        X = check_array(X, dtype='float64')
        scaler = SklearnMaxAbsScaler()
        X_new = scaler.fit_transform(X.T).T
        return X_new


class RobustScaler(BaseEstimator, TransformerMixin):
    """Scale samples using statistics that are robust to outliers.

    This Scaler removes the median and scales the data according to
    the quantile range (defaults to IQR: Interquartile Range).
    The IQR is the range between the 1st quartile (25th quantile)
    and the 3rd quartile (75th quantile).

    Centering and scaling happen independently on each sample by
    computing the relevant statistics on the timestamps. Median and
    interquartile range are used to transform the data.

    Standardization of a dataset is a common requirement for many
    machine learning estimators. Typically this is done by removing the mean
    and scaling to unit variance. However, outliers can often influence the
    sample mean / variance in a negative way. In such cases, the median and
    the interquartile range often give better results.

    Parameters
    ----------
    with_centering : bool (default = True)
        If True, center the data before scaling.

    with_scaling : bool (default = True)
        If True, scale the data to interquartile range.

    quantile_range : tuple (q_min, q_max), 0.0 < q_min < q_max < 100.0
        Default: (25.0, 75.0) = (1st quantile, 3rd quantile) = IQR

    Examples
    --------
    >>> from pyts.preprocessing import RobustScaler
    >>> X = [[1, -2,  4],
    ...      [-2,  1,  1],
    ...      [2,  3, -2]]
    >>> scaler = RobustScaler()
    >>> scaler.transform(X)
    array([[ 0. , -1. ,  1. ],
           [-2. ,  0. ,  0. ],
           [ 0. ,  0.4, -1.6]])

    """

    def __init__(self, with_centering=True, with_scaling=True,
                 quantile_range=(25.0, 75.0)):
        self.with_centering = with_centering
        self.with_scaling = with_scaling
        self.quantile_range = quantile_range

    def fit(self, X=None, y=None):
        """Pass.

        Parameters
        ----------
        X
            Ignored

        y
            Ignored

        Returns
        -------
        self : object

        """
        return self

    def transform(self, X):
        """Scale the data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Data to scale.

        Returns
        -------
        X_new : array-like, shape = (n_samples, n_timestamps)
            Scaled data.

        """
        X = check_array(X, dtype='float64')
        scaler = SklearnRobustScaler(
            with_centering=self.with_centering,
            with_scaling=self.with_scaling,
            quantile_range=self.quantile_range
        )
        X_new = scaler.fit_transform(X.T).T
        return X_new



"""Utility class for multivariate time series classification."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
from numba import njit, prange
from sklearn.base import BaseEstimator, clone
from sklearn.preprocessing import LabelEncoder
from sklearn.utils.validation import check_is_fitted
from ...base import MultivariateClassifierMixin
from ..utils import check_3d_array


@njit()
def _hard_vote(y_pred, weights):
    n_samples, n_features = y_pred.shape
    maj = np.empty(n_samples, dtype=np.int64)
    for i in prange(n_samples):
        maj[i] = np.argmax(np.bincount(y_pred[i], weights))
    return maj


class MultivariateClassifier(BaseEstimator, MultivariateClassifierMixin):
    """Classifier for multivariate time series.

    It provides a convenient class to classify multivariate time series with
    classifier that can only deal with univariate time series. The labels are
    predicted in a hard voting fashion using the predictions for each feature.

    Parameters
    ----------
    estimator : estimator object or list thereof
        Classifier. If one estimator is provided, it is cloned and each clone
        performs prediction for one feature. If a list of estimators is
        provided, each estimator performs prediction for one feature.

    weights : array-like, shape = (n_classifiers,) or None (default=None)
        Sequence of weights (`float` or `int`) to weight the occurrences of
        predicted class labels. Uses uniform weights if None.

    Attributes
    ----------
    classes_ : array, shape = (n_classes,)
        An array of class labels known to the classifier.

    estimators_ : list of estimator objects
        The collection of fitted classifiers.

    Examples
    --------
    >>> from pyts.classification import BOSSVS
    >>> from pyts.datasets import load_basic_motions
    >>> from pyts.multivariate.classification import MultivariateClassifier
    >>> X_train, X_test, y_train, y_test = load_basic_motions(return_X_y=True)
    >>> clf = MultivariateClassifier(BOSSVS())
    >>> clf.fit(X_train, y_train)
    MultivariateClassifier(...)
    >>> clf.score(X_test, y_test)
    1.0

    """

    def __init__(self, estimator, weights=None):
        self.estimator = estimator
        self.weights = weights

    def fit(self, X, y):
        """Fit each classifier.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features, n_timestamps)
            Multivariate time series.

        y : None or array-like, shape = (n_samples,)
            Class labels.

        Returns
        -------
        self : object

        """
        X = check_3d_array(X)
        _, n_features, _ = X.shape
        self._check_params(n_features)

        if self.weights is None:
            self._weights = None
        else:
            self._weights = np.asarray(self.weights)

        self._le = LabelEncoder().fit(y)
        self.classes_ = self._le.classes_
        y_ind = self._le.transform(y)
        for i, clf in enumerate(self.estimators_):
            clf.fit(X[:, i, :], y_ind)
        return self

    def predict(self, X):
        """Predict class labels using hard voting.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features, n_timestamps)
            Multivariate time series.

        Returns
        -------
        y_pred : array, shape = (n_samples,)
            Predicted class labels.

        """
        X = check_3d_array(X)
        check_is_fitted(self, 'estimators_')
        n_samples, n_features, _ = X.shape

        y_pred = np.empty((n_samples, n_features))
        for i, clf in enumerate(self.estimators_):
            y_pred[:, i] = clf.predict(X[:, i, :])

        maj = _hard_vote(y_pred.astype('int64'), self._weights)
        return self._le.inverse_transform(maj)

    def _check_params(self, n_features):
        """Check parameters."""
        classifier = (isinstance(self.estimator, BaseEstimator)
                      and hasattr(self.estimator, 'predict'))
        if classifier:
            self.estimators_ = [clone(self.estimator)
                                for _ in range(n_features)]

        elif isinstance(self.estimator, list):
            if len(self.estimator) != n_features:
                raise ValueError(
                    "If 'estimator' is a list, its length must be equal to "
                    "the number of features ({0} != {1})"
                    .format(len(self.estimator), n_features)
                )
            for i, estimator in enumerate(self.estimator):
                if not (isinstance(estimator, BaseEstimator)
                        and hasattr(estimator, 'predict')):
                    raise ValueError("Estimator {} must be a classifier."
                                     .format(i))
            self.estimators_ = self.estimator

        else:
            raise TypeError(
                "'estimator' must be a classifier that inherits from "
                "sklearn.base.BaseEstimator or a list thereof."
            )



"""The :mod:`pyts.transformation` module includes transformation algorithms."""

from .bag_of_patterns import BagOfPatterns
from .boss import BOSS
from .rocket import ROCKET
from .shapelet_transform import ShapeletTransform
from .weasel import WEASEL

__all__ = ['BagOfPatterns', 'BOSS', 'ROCKET', 'ShapeletTransform', 'WEASEL']
